import numpy as np
exp = np.exp


"""Dynamics description and simulation"""


def unspliced(tau, u0, alpha, beta):
    expu = exp(-beta * tau)
    return u0 * expu + alpha / beta * (1 - expu)


def spliced(tau, s0, u0, alpha, beta, gamma):
    c = (alpha - u0 * beta) / (gamma - beta)
    expu, exps = exp(-beta * tau), exp(-gamma * tau)
    return s0 * exps + alpha / gamma * (1 - exps) + c * (exps - expu)


def vectorize(t, t_, alpha, beta, gamma, alpha_=0, u0=0, s0=0):
    o = np.array(t < t_, dtype=bool)
    tau = t * o + (t - t_) * (1 - o)

    u0_ = unspliced(t_, u0, alpha, beta)
    s0_ = spliced(t_, s0, u0, alpha, beta, gamma)

    # vectorize u0, s0 and alpha
    u0 = u0 * o + u0_ * (1 - o)
    s0 = s0 * o + s0_ * (1 - o)
    alpha = alpha * o + alpha_ * (1 - o)

    return tau, alpha, u0, s0


"""State-independent derivatives"""


def du_du0(beta, tau):
    return exp(-beta * tau)


def ds_ds0(gamma, tau):
    return exp(-gamma * tau)


def ds_du0(beta, gamma, tau):
    return - beta / (gamma - beta) * (exp(-gamma * tau) - exp(-beta * tau))


def du_dalpha(beta, tau):
    return 1 / beta * (1 - exp(-beta * tau))


def ds_dalpha(beta, gamma, tau):
    expu, exps = exp(-beta * tau), exp(-gamma * tau)
    return 1 / gamma * (1 - exps) + 1 / (gamma - beta) * (exps - expu)


def du_dbeta(alpha, beta, tau, u0=0):
    expu = exp(-beta * tau)
    return (alpha / beta - u0) * tau * expu - alpha / beta ** 2 * (1 - expu)


def ds_dbeta(alpha, beta, gamma, tau, u0=0):
    cu = (alpha - beta * u0) / (gamma - beta)
    cs = (alpha - gamma * u0) / (gamma - beta) ** 2
    expu, exps = exp(-beta * tau), exp(-gamma * tau)
    return cu * tau * expu + cs * (exps - expu)


def ds_dgamma(alpha, beta, gamma, tau, u0=0, s0=0):
    cu = (alpha - beta * u0) / (gamma - beta)
    cs = alpha / gamma - s0
    expu, exps = exp(-beta * tau), exp(-gamma * tau)
    return (cs - cu) * tau * exps - alpha / gamma ** 2 * (1 - exps) - cu / (gamma - beta) * (exps - expu)


def derivatives(alpha, beta, gamma, tau, o, diffU, diffS, alpha_=0, u0=0, s0=0, u0_=None, s0_=None, weights=None):
    # state-dependent derivatives:
    du_u = du_du0(beta, tau)
    ds_u = ds_du0(beta, gamma, tau)
    dL_du0 = (du_u * o).dot(diffU) + (ds_u * o).dot(diffS)
    dL_du0_ = (du_u * (1 - o)).dot(diffU) + (ds_u * (1 - o)).dot(diffS)

    ds_s = ds_ds0(gamma, tau)
    dL_ds0 = (ds_s * o).dot(diffS)
    dL_ds0_ = (ds_s * (1 - o)).dot(diffS)

    du_a = du_dalpha(beta, tau)
    ds_a = ds_dalpha(beta, gamma, tau)
    dL_dalpha = (du_a * o).dot(diffU) + (ds_a * o).dot(diffS)
    dL_dalpha_ = (du_a * (1 - o)).dot(diffU) + (ds_a * (1 - o)).dot(diffS)

    # independent derivatives: beta, gamma
    if u0_ is None: u0_ = alpha / beta
    if s0_ is None: s0_ = alpha / gamma

    u0 = u0 * o + u0_ * (1 - o)
    s0 = s0 * o + s0_ * (1 - o)

    alpha = alpha * o + alpha_ * (1 - o)

    dL_dbeta = du_dbeta(alpha, beta, tau, u0).dot(diffU) + ds_dbeta(alpha, beta, gamma, tau, u0).dot(diffS)
    dL_dgamma = ds_dgamma(alpha, beta, gamma, tau, u0, s0).dot(diffS)

    return dL_dalpha, dL_dbeta, dL_dgamma, dL_dalpha_, dL_du0, dL_ds0, dL_du0_, dL_ds0_


"""State-independent derivatives"""


def du_dalpha_o(beta, tau, state=1):
    expU = np.exp(-beta * tau)
    _expU = (1 - expU) * (state == 1) + expU * (state == 0)
    return 1 / beta * _expU


def du_dbeta_o(alpha, beta, tau, state=1, u0=0):
    expU = np.exp(-beta * tau)
    _expU = (1 - expU) * (state == 1) + expU * (state == 0)
    c = (alpha / beta - u0) * (state == 1) - alpha / beta * (state == 0)
    return c * tau * expU - alpha / beta ** 2 * _expU


def ds_dalpha_o(beta, gamma, tau, state=1):
    expU, expS = np.exp(-beta * tau), np.exp(-gamma * tau)
    _expU = (1 - expU) * (state == 1) + expU * (state == 0)
    _expS = (1 - expS) * (state == 1) + expS * (state == 0)
    return 1 / gamma * _expS - 1 / (gamma - beta) * (_expS - _expU)


def ds_dbeta_o(alpha, beta, gamma, tau, state=1, u0=0):
    _cS = (alpha - gamma * u0) / (gamma - beta) ** 2 * (state == 1) - alpha / (gamma - beta) ** 2 * (state == 0)
    _cU = (alpha - beta * u0) / (gamma - beta) * (state == 1) - alpha / (gamma - beta) * (state == 0)
    expU, expS = np.exp(-beta * tau), np.exp(-gamma * tau)
    return _cU * tau * expU + _cS * (expS - expU)


def ds_dgamma_o(alpha, beta, gamma, tau, state=1, u0=0, s0=0):
    _cU = (alpha - beta * u0) / (gamma - beta) * (state == 1) - alpha / (gamma - beta) * (state == 0)
    _cS = alpha / gamma - s0 * (state == 1) - alpha / gamma * (state == 0)
    expU, expS = np.exp(-beta * tau), np.exp(-gamma * tau)
    _expS = (1 - expS) * (state == 1) + expS * (state == 0)
    return (_cS - _cU) * tau * expS - alpha / gamma ** 2 * _expS - _cU / (gamma - beta) * (expS - expU)


def derivatives_o(alpha, beta, gamma, tau, o, diffU, diffS, alpha_=0, u0=0, s0=0, u0_=None, s0_=None):
    dL_dalpha = du_dalpha_o(beta, tau, o).dot(diffU) + ds_dalpha_o(beta, gamma, tau, o).dot(diffS)
    dL_dbeta = du_dbeta_o(alpha, beta, tau, o).dot(diffU) + ds_dbeta_o(alpha, beta, gamma, tau, o).dot(diffS)
    dL_dgamma = ds_dgamma_o(alpha, beta, gamma, tau, o).dot(diffS)

    return dL_dalpha, dL_dbeta, dL_dgamma, 0, 0, 0, 0, 0



from .. import settings
from .. import logging as logg
from ..preprocessing.moments import get_connectivities
from .velocity_graph import VelocityGraph
from .transition_matrix import transition_matrix
from .utils import scale, groups_to_bool, strings_to_categoricals

from scipy.sparse import linalg, csr_matrix, issparse
import numpy as np


def cell_fate(data, groupby='clusters', disconnected_groups=None, n_neighbors=None, copy=False):
    """Computes individual cell endpoints

    Arguments
    ---------
    data: :class:`~anndata.AnnData`
        Annotated data matrix.
    groupby: `str` (default: `'clusters'`)
        Key to which to assign the fates.
    disconnected_groups: list of `str` (default: `None`)
        Which groups to treat as disconnected for fate assignment.
    n_neighbors: `int` (default: `None`)
        Number of neighbors to restrict transitions to.
    copy: `bool` (default: `False`)
        Return a copy instead of writing to `adata`.

    Returns
    -------
    Returns or updates `adata` with the attributes
    cell_fate: `.obs`
        most likely cell fate for each individual cell
    cell_fate_confidence: `.obs`
        confidence of transitioning to the assigned fate
    """
    adata = data.copy() if copy else data
    logg.info('computing cell fates', r=True)

    n_neighbors = 10 if n_neighbors is None else n_neighbors
    _adata = adata.copy()
    vgraph = VelocityGraph(_adata, n_neighbors=n_neighbors, approx=True, n_recurse_neighbors=1)
    vgraph.compute_cosines()
    _adata.uns['velocity_graph'] = vgraph.graph
    _adata.uns['velocity_graph_neg'] = vgraph.graph_neg

    T = transition_matrix(_adata)
    I = np.eye(_adata.n_obs)
    fate = np.linalg.inv(I - T)
    if issparse(T): fate = fate.A
    cell_fates = np.array(_adata.obs[groupby][fate.argmax(1)])
    if disconnected_groups is not None:
        idx = _adata.obs[groupby].isin(disconnected_groups)
        cell_fates[idx] = _adata.obs[groupby][idx]
    adata.obs['cell_fate'] = cell_fates
    adata.obs['cell_fate_confidence'] = fate.max(1) / fate.sum(1)
    strings_to_categoricals(adata)

    logg.info('    finished', time=True, end=' ' if settings.verbosity > 2 else '\n')
    logg.hint(
        'added\n'
        '    \'cell_fate\', most likely cell fate (adata.obs)\n'
        '    \'cell_fate_confidence\', confidence of transitioning to the assigned fate (adata.obs)')
    return adata if copy else None


def cell_origin(data, groupby='clusters', distinct_clusters=None, self_transitions=False, n_neighbors=None):
    n_neighbors = 10 if n_neighbors is None else n_neighbors
    adata = data.copy()
    vgraph = VelocityGraph(adata, n_neighbors=n_neighbors, approx=True, n_recurse_neighbors=1)
    vgraph.compute_cosines()
    adata.uns['velocity_graph'] = vgraph.graph
    adata.uns['velocity_graph_neg'] = vgraph.graph_neg

    T = transition_matrix(adata, self_transitions=self_transitions, backward=True)
    I = np.eye(data.n_obs)
    fate = np.linalg.inv(I - T)
    if issparse(T): fate = fate.A
    cell_fates = np.array(data.obs[groupby][fate.argmax(1)])
    if distinct_clusters is not None:
        idx = data.obs[groupby].isin(distinct_clusters)
        cell_fates[idx] = data.obs[groupby][idx]
    data.obs['cell_origin'] = cell_fates
    data.obs['cell_origin_confidence'] = fate.max(1) / fate.sum(1)
    strings_to_categoricals(data)


def eigs(T, k=10, eps=1e-3, perc=None):
    try:
        eigvals, eigvecs = linalg.eigs(T.T, k=k, which='LR')  # find k eigs with largest real part

        p = np.argsort(eigvals)[::-1]                        # sort in descending order of eigenvalues
        eigvals = eigvals.real[p]
        eigvecs = eigvecs.real[:, p]

        idx = (eigvals >= 1 - eps)                           # select eigenvectors with eigenvalue of 1
        eigvals = eigvals[idx]
        eigvecs = np.absolute(eigvecs[:, idx])

        if perc is not None:
            lbs, ubs = np.percentile(eigvecs, perc, axis=0)
            eigvecs[eigvecs < lbs] = 0
            eigvecs = np.clip(eigvecs, 0, ubs)
            eigvecs /= eigvecs.max(0)

    except:
        eigvals, eigvecs = np.empty(0), np.zeros(shape=(T.shape[0], 0))

    return eigvals, eigvecs


def write_to_obs(adata, key, vals, cell_subset=None):
    if cell_subset is None:
        adata.obs[key] = vals
    else:
        vals_all = adata.obs[key].copy() if key in adata.obs.keys() else np.zeros(adata.n_obs)
        vals_all[cell_subset] = vals
        adata.obs[key] = vals_all


def terminal_states(data, vkey='velocity', groupby=None, groups=None, self_transitions=False, basis=None,
                    weight_diffusion=0, scale_diffusion=1, eps=1e-3, copy=False):
    """Computes terminal states (root and end points).

    Arguments
    ---------
    data: :class:`~anndata.AnnData`
        Annotated data matrix.
    vkey: `str` (default: `'velocity'`)
        Name of velocity estimates to be used.
    self_transitions: `bool` (default: `False`)
        Allow transitions from one node to itself.
    basis: `str` (default: `None`)
        Basis to use.
    weight_diffusion: `float` (default: 0)
        Relative weight to be given to diffusion kernel (Brownian motion)
    scale_diffusion: `float` (default: 1)
        Scale of diffusion kernel.
    eps: `float` (default: 1e-3)
        Tolerance for eigenvalue selection.
    copy: `bool` (default: `False`)
        Return a copy instead of writing to data.

    Returns
    -------
    Returns or updates `data` with the attributes
    root: `.obs`
        sparse matrix with transition probabilities.
    end: `.obs`
        sparse matrix with transition probabilities.
    """
    adata = data.copy() if copy else data
    logg.info('computing terminal states', r=True)

    strings_to_categoricals(adata)

    groupby = 'cell_fate' if groupby is None and 'cell_fate' in adata.obs.keys() else groupby
    categories = adata.obs[groupby].cat.categories if groupby is not None and groups is None else [None]
    for cat in categories:
        groups = cat if cat is not None else groups
        cell_subset = groups_to_bool(adata, groups=groups, groupby=groupby)
        _adata = adata if groups is None else adata[cell_subset]
        connectivities = get_connectivities(_adata, 'distances')

        T = transition_matrix(_adata, vkey=vkey, basis=basis, weight_diffusion=weight_diffusion,
                              scale_diffusion=scale_diffusion, self_transitions=self_transitions, backward=True)
        eigvecs_roots = eigs(T, eps=eps, perc=[2, 98])[1]
        roots = csr_matrix.dot(connectivities, eigvecs_roots).sum(1)
        roots = scale(np.clip(roots, 0, np.percentile(roots, 98)))
        write_to_obs(adata, 'root_cells', roots, cell_subset)

        T = transition_matrix(_adata, vkey=vkey, basis=basis, weight_diffusion=weight_diffusion,
                              scale_diffusion=scale_diffusion, self_transitions=self_transitions, backward=False)
        eigvecs_ends = eigs(T, eps=eps, perc=[2, 98])[1]
        ends = csr_matrix.dot(connectivities, eigvecs_ends).sum(1)
        ends = scale(np.clip(ends, 0, np.percentile(ends, 98)))
        write_to_obs(adata, 'end_points', ends, cell_subset)

        n_roots, n_ends = eigvecs_roots.shape[1], eigvecs_ends.shape[1]
        groups_str = ' (' + groups + ')' if isinstance(groups, str) else ''
        logg.info('    identified ' + str(n_roots) + ' root cells and ' + str(n_ends) + ' end points' + groups_str)

    logg.info('    finished', time=True, end=' ' if settings.verbosity > 2 else '\n')
    logg.hint(
        'added\n'
        '    \'root_cells\', root cells of Markov diffusion process (adata.obs)\n'
        '    \'end_points\', end points of Markov diffusion process (adata.obs)')
    return adata if copy else None



from ..tools.velocity_embedding import velocity_embedding
from ..tools.utils import groups_to_bool
from .utils import default_basis, default_size, default_color, get_components, savefig_or_show, make_unique_list
from .velocity_embedding_grid import compute_velocity_on_grid
from .scatter import scatter
from .docs import doc_scatter, doc_params

from matplotlib import rcParams
import matplotlib.pyplot as pl
import numpy as np


@doc_params(scatter=doc_scatter)
def velocity_embedding_stream(adata, basis=None, vkey='velocity', density=None, smooth=None, linewidth=None,
                              n_neighbors=None, X=None, V=None, X_grid=None, V_grid=None, color=None, use_raw=None,
                              layer=None, color_map=None, colorbar=True, palette=None, size=None, alpha=.1, perc=None,
                              sort_order=True, groups=None, components=None, legend_loc='on data',
                              legend_fontsize=None, legend_fontweight=None, right_margin=None, left_margin=None,
                              xlabel=None, ylabel=None, title=None, fontsize=None, figsize=None, dpi=None, frameon=None,
                              show=True, save=None, ax=None, ncols=None, **kwargs):
    """\
    Stream plot of velocities on the embedding.

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    x: `str`, `np.ndarray` or `None` (default: `None`)
        x coordinate
    y: `str`, `np.ndarray` or `None` (default: `None`)
        y coordinate
    vkey: `str` or `None` (default: `None`)
        Key for annotations of observations/cells or variables/genes.
    density: `float` (default: 1)
        Amount of velocities to show - 0 none to 1 all
    smooth: `float` (default: 0.5)
        Multiplication factor for scale in Gaussian kernel around grid point.
    linewidth: `float` (default: 1)
        Line width for streamplot.
    n_neighbors: `int` (default: None)
        Number of neighbors to consider around grid point.
    X: `np.ndarray` (default: None)
        Embedding grid point coordinates
    V: `np.ndarray` (default: None)
        Embedding grid velocity coordinates
    {scatter}

    Returns
    -------
        `matplotlib.Axis` if `show==False`
    """
    basis = default_basis(adata) if basis is None else basis
    colors, layers, vkeys = make_unique_list(color, allow_array=True), make_unique_list(layer), make_unique_list(vkey)
    for key in vkeys:
        if key + '_' + basis not in adata.obsm_keys() and V is None:
            velocity_embedding(adata, basis=basis, vkey=key)
    color, layer, vkey = colors[0], layers[0], vkeys[0]
    color = default_color(adata) if color is None else color

    if X_grid is None or V_grid is None:
        _adata = adata[groups_to_bool(adata, groups, groupby=color)] \
            if groups is not None and color in adata.obs.keys() else adata
        X_emb = _adata.obsm['X_' + basis][:, get_components(components, basis)] if X is None else X[:, :2]
        V_emb = _adata.obsm[vkey + '_' + basis][:, get_components(components, basis)] if V is None else V[:, :2]
        X_grid, V_grid = compute_velocity_on_grid(X_emb=X_emb, V_emb=V_emb, density=1, smooth=smooth,
                                                  n_neighbors=n_neighbors, autoscale=False, adjust_for_stream=True)
        lengths = np.sqrt((V_grid ** 2).sum(0))
        linewidth = 1 if linewidth is None else linewidth
        linewidth *= 2 * lengths / lengths[~np.isnan(lengths)].max()

    scatter_kwargs = {"basis": basis, "perc": perc, "use_raw": use_raw, "sort_order": sort_order, "alpha": alpha,
                      "components": components, "legend_loc": legend_loc, "groups": groups,
                      "legend_fontsize": legend_fontsize, "legend_fontweight": legend_fontweight, "palette": palette,
                      "color_map": color_map, "frameon": frameon, "xlabel": xlabel, "ylabel": ylabel,
                      "right_margin": right_margin, "left_margin": left_margin, "colorbar": colorbar, "dpi": dpi,
                      "fontsize": fontsize, "show": False, "save": None}

    multikey = colors if len(colors) > 1 else layers if len(layers) > 1 else vkeys if len(vkeys) > 1 else None
    if multikey is not None:
        if isinstance(title, (list, tuple)): title *= int(np.ceil(len(multikey) / len(title)))
        ncols = len(multikey) if ncols is None else min(len(multikey), ncols)
        nrows = int(np.ceil(len(multikey) / ncols))
        figsize = rcParams['figure.figsize'] if figsize is None else figsize
        for i, gs in enumerate(
                pl.GridSpec(nrows, ncols, pl.figure(None, (figsize[0] * ncols, figsize[1] * nrows), dpi=dpi))):
            if i < len(multikey):
                velocity_embedding_stream(adata, density=density, size=size, smooth=smooth, n_neighbors=n_neighbors,
                                          linewidth=linewidth, ax=pl.subplot(gs),
                                          color=colors[i] if len(colors) > 1 else color,
                                          layer=layers[i] if len(layers) > 1 else layer,
                                          vkey=vkeys[i] if len(vkeys) > 1 else vkey,
                                          title=title[i] if isinstance(title, (list, tuple)) else title,
                                          X_grid=None if len(vkeys) > 1 else X_grid,
                                          V_grid=None if len(vkeys) > 1 else V_grid, **scatter_kwargs, **kwargs)
        savefig_or_show('' if basis is None else basis, dpi=dpi, save=save, show=show)
        if not show: return ax

    else:
        ax = pl.figure(None, figsize, dpi=dpi).gca() if ax is None else ax

        density = 1 if density is None else density
        stream_kwargs = {"linewidth": linewidth, "density": 2 * density}
        stream_kwargs.update(kwargs)
        pl.streamplot(X_grid[0], X_grid[1], V_grid[0], V_grid[1], color='grey', zorder=3, **stream_kwargs)

        size = 4 * default_size(adata) if size is None else size
        ax = scatter(adata, layer=layer, color=color, size=size, title=title, ax=ax, zorder=0, **scatter_kwargs)

        savefig_or_show('' if basis is None else basis, dpi=dpi, save=save, show=show)
        if not show: return ax



from .. import logging as logg

import numpy as np
from scipy.sparse import issparse
from sklearn.utils import sparsefuncs
from anndata import AnnData


def show_proportions(adata):
    """Fraction of spliced/unspliced/ambiguous abundances

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.

    Returns
    -------
    Prints the fractions of abundances.
    """
    layers_keys = [key for key in ['spliced', 'unspliced', 'ambiguous'] if key in adata.layers.keys()]
    tot_mol_cell_layers = [adata.layers[key].sum(1) for key in layers_keys]

    mean_abundances = np.round(
        [np.mean(tot_mol_cell / np.sum(tot_mol_cell_layers, 0)) for tot_mol_cell in tot_mol_cell_layers], 2)

    print('Abundance of ' + str(layers_keys) + ': ' + str(mean_abundances))


def cleanup(data, clean='layers', keep=None, copy=False):
    """Deletes attributes not needed.

    Arguments
    ---------
    data: :class:`~anndata.AnnData`
        Annotated data matrix.
    clean: `str` or list of `str` (default: `layers`)
        Which attributes to consider for freeing memory.
    keep: `str` or list of `str` (default: `['spliced', unspliced']`)
        Which attributes to keep.
    copy: `bool` (default: `False`)
        Return a copy instead of writing to adata.

    Returns
    -------
    Returns or updates `adata` with selection of attributes kept.
    """
    adata = data.copy() if copy else data

    keep = list() if keep is None else list(keep)
    keep.extend(['spliced', 'unspliced', 'Ms', 'Mu', 'clusters', 'neighbors'])

    if any(['obs' in clean, 'all' in clean]):
        for key in list(adata.obs.keys()):
            if key not in keep: del adata.obs[key]

    if any(['var' in clean, 'all' in clean]):
        for key in list(adata.var.keys()):
            if key not in keep: del adata.var[key]

    if any(['uns' in clean, 'all' in clean]):
        for key in list(adata.uns.keys()):
            if key not in keep: del adata.uns[key]

    if any(['layers' in clean, 'all' in clean]):
        for key in list(adata.layers.keys()):  # remove layers that are not needed
            if key not in keep: del adata.layers[key]

    return adata if copy else None


def get_size(adata, layer=None):
    X = adata.X if layer is None else adata.layers[layer]
    return X.sum(1).A1 if issparse(X) else X.sum(1)


def set_initial_size(adata, layers={'spliced', 'unspliced'}):
    layers = [layer for layer in layers
              if layer in adata.layers.keys() and 'initial_size_' + layer not in adata.obs.keys()]
    for layer in layers:
        adata.obs['initial_size_' + layer] = get_size(adata, layer)
    if 'initial_size' not in adata.obs.keys(): adata.obs['initial_size'] = get_size(adata)


def get_initial_size(adata, layer=None, by_total_size=None):
    if by_total_size:
        return np.sum([adata.obs['initial_size_' + layer] for layer in {'spliced', 'unspliced'}
                       if 'initial_size_' + layer in adata.obs.keys()], axis=0)
    elif layer in adata.layers.keys():
        return np.array(adata.obs['initial_size_' + layer]) if 'initial_size_' + layer in adata.obs.keys() else get_size(adata, layer)
    elif layer is None or layer is 'X':
        return np.array(adata.obs['initial_size']) if 'initial_size' in adata.obs.keys() else get_size(adata)
    else:
        return None


def filter(X, min_counts=None, min_cells=None, max_counts=None, max_cells=None):
    counts = X if (min_counts is not None or max_counts is not None) else X > 0
    counts = counts.sum(0).A1 if issparse(counts) else counts.sum(0)
    lb = min_counts if min_counts is not None else min_cells if min_cells is not None else -np.inf
    ub = max_counts if max_counts is not None else max_cells if max_cells is not None else np.inf
    return (lb <= counts) & (counts <= ub), counts


def filter_genes(data, min_counts=None, min_cells=None, max_counts=None, max_cells=None,
                 min_counts_u=None,  min_cells_u=None, max_counts_u=None, max_cells_u=None, copy=False):
    """Filter genes based on number of cells or counts.
    Keep genes that have at least `min_counts` counts or are expressed in at
    least `min_cells` cells or have at most `max_counts` counts or are expressed
    in at most `max_cells` cells.
    Only provide one of the optional parameters `min_counts`, `min_cells`,
    `max_counts`, `max_cells` per call.

    Parameters
    ----------
    data : :class:`~anndata.AnnData`, `np.ndarray`, `sp.spmatrix`
        The (annotated) data matrix of shape `n_obs` × `n_vars`. Rows correspond
        to cells and columns to genes.
    min_counts : `int`, optional (default: `None`)
        Minimum number of counts required for a gene to pass filtering.
    min_cells : `int`, optional (default: `None`)
        Minimum number of cells expressed required for a gene to pass filtering.
    max_counts : `int`, optional (default: `None`)
        Maximum number of counts required for a gene to pass filtering.
    max_cells : `int`, optional (default: `None`)
        Maximum number of cells expressed required for a gene to pass filtering.
    min_counts_u : `int`, optional (default: `None`)
        Minimum number of unspliced counts required for a gene to pass filtering.
    min_cells_u : `int`, optional (default: `None`)
        Minimum number of unspliced cells expressed required for a gene to pass filtering.
    max_counts_u : `int`, optional (default: `None`)
        Maximum number of unspliced counts required for a gene to pass filtering.
    max_cells_u : `int`, optional (default: `None`)
        Maximum number of unspliced cells expressed required for a gene to pass filtering.
    copy : `bool`, optional (default: `False`)
        Determines whether a copy is returned.

    Returns
    -------
    Filters the object and adds `n_counts` to `adata.var`.
    """
    adata = data.copy() if copy else data

    # set initial cell sizes before filtering
    set_initial_size(adata)

    layers = [layer for layer in ['spliced', 'unspliced'] if layer in adata.layers.keys()]
    for layer in layers:
        X = adata.layers[layer]
        if layer is 'spliced':
            _min_counts, _min_cells, _max_counts, _max_cells = min_counts, min_cells, max_counts, max_cells
        else:
            _min_counts, _min_cells, _max_counts, _max_cells = min_counts_u, min_cells_u, max_counts_u, max_cells_u

        if layer in adata.layers.keys():
            gene_subset = np.ones(adata.n_vars, dtype=bool)
            if _min_counts is not None or _max_counts is not None:
                gene_subset, _ = filter(X, min_counts=_min_counts, max_counts=_max_counts)
                adata._inplace_subset_var(gene_subset)

            if _min_cells is not None or _max_cells is not None:
                gene_subset, _ = filter(X, min_cells=_min_cells, max_cells=_max_cells)
                adata._inplace_subset_var(gene_subset)

            s = np.sum(~gene_subset)
            if s > 0:
                logg.info('Filtered out {} genes that are detected'.format(s), end=' ')
                if _min_cells is not None or _min_counts is not None:
                    logg.info('in less than', str(_min_cells) + ' cells (' + str(layer) + ').' if _min_counts is None
                              else str(_min_counts) + ' counts (' + str(layer) + ').', no_indent=True)
                if max_cells is not None or max_counts is not None:
                    logg.info('in more than ', str(_max_cells) + ' cells(' + str(layer) + ').' if _max_counts is None
                              else str(_max_counts) + ' counts (' + str(layer) + ').', no_indent=True)

    return adata if copy else None


def filter_genes_dispersion(data, flavor='seurat', min_disp=None, max_disp=None, min_mean=None, max_mean=None,
                            n_bins=20, n_top_genes=None, log=True, copy=False):
    """Extract highly variable genes.
    The normalized dispersion is obtained by scaling with the mean and standard
    deviation of the dispersions for genes falling into a given bin for mean
    expression of genes. This means that for each bin of mean expression, highly
    variable genes are selected.

    Parameters
    ----------
    data : :class:`~anndata.AnnData`, `np.ndarray`, `sp.sparse`
        The (annotated) data matrix of shape `n_obs` × `n_vars`. Rows correspond
        to cells and columns to genes.
    flavor : {'seurat', 'cell_ranger', 'svr'}, optional (default: 'seurat')
        Choose the flavor for computing normalized dispersion. If choosing
        'seurat', this expects non-logarithmized data - the logarithm of mean
        and dispersion is taken internally when `log` is at its default value
        `True`. For 'cell_ranger', this is usually called for logarithmized data
        - in this case you should set `log` to `False`. In their default
        workflows, Seurat passes the cutoffs whereas Cell Ranger passes
        `n_top_genes`.
    min_mean=0.0125, max_mean=3, min_disp=0.5, max_disp=`None` : `float`, optional
        If `n_top_genes` unequals `None`, these cutoffs for the means and the
        normalized dispersions are ignored.
    n_bins : `int` (default: 20)
        Number of bins for binning the mean gene expression. Normalization is
        done with respect to each bin. If just a single gene falls into a bin,
        the normalized dispersion is artificially set to 1. You'll be informed
        about this if you set `settings.verbosity = 4`.
    n_top_genes : `int` or `None` (default: `None`)
        Number of highly-variable genes to keep.
    log : `bool`, optional (default: `True`)
        Use the logarithm of the mean to variance ratio.
    copy : `bool`, optional (default: `False`)
        If an :class:`~anndata.AnnData` is passed, determines whether a copy
        is returned.

    Returns
    -------
    If an AnnData `adata` is passed, returns or updates `adata` depending on \
    `copy`. It filters the `adata` and adds the annotations
    """
    adata = data.copy() if copy else data
    set_initial_size(adata)
    if n_top_genes is not None and adata.n_vars < n_top_genes:
        logg.info('Skip filtering by dispersion since number of variables are less than `n_top_genes`')
    else:
        if flavor is 'svr':
            mu = adata.X.mean(0).A1 if issparse(adata.X) else adata.X.mean(0)
            sigma = np.sqrt(adata.X.multiply(adata.X).mean(0).A1 - mu ** 2) if issparse(adata.X) else adata.X.std(0)
            log_mu = np.log2(mu)
            log_cv = np.log2(sigma / mu)

            from sklearn.svm import SVR
            clf = SVR(gamma=150. / len(mu))
            clf.fit(log_mu[:, None], log_cv)
            score = log_cv - clf.predict(log_mu[:, None])
            nth_score = np.sort(score)[::-1][n_top_genes]
            adata._inplace_subset_var(score >= nth_score)
        else:
            from scanpy.api.pp import filter_genes_dispersion
            filter_genes_dispersion(adata, flavor=flavor, min_disp=min_disp, max_disp=max_disp, min_mean=min_mean,
                                    max_mean=max_mean, n_bins=n_bins, n_top_genes=n_top_genes, log=log)
    return adata if copy else None


def counts_per_cell_quantile(X, max_proportion_per_cell=.05, counts_per_cell=None):
    if counts_per_cell is None:
        counts_per_cell = X.sum(1).A1 if issparse(X) else X.sum(1)
    gene_subset = np.all(X <= counts_per_cell[:, None] * max_proportion_per_cell, axis=0)
    if issparse(X): gene_subset = gene_subset.A1
    return X[:, gene_subset].sum(1).A1 if issparse(X) else X[:, gene_subset].sum(1)


def not_yet_normalized(X):
    return np.allclose((X.data[:10] if issparse(X) else X[:, 0]) % 1, 0, atol=1e-3)


def normalize_per_cell(data, counts_per_cell_after=None, counts_per_cell=None, key_n_counts=None,
                       max_proportion_per_cell=None, use_initial_size=True, layers=['spliced', 'unspliced'],
                       enforce=False, copy=False):
    """Normalize each cell by total counts over all genes.

    Parameters
    ----------
    data : :class:`~anndata.AnnData`, `np.ndarray`, `sp.sparse`
        The (annotated) data matrix of shape `n_obs` × `n_vars`. Rows correspond
        to cells and columns to genes.
    counts_per_cell_after : `float` or `None`, optional (default: `None`)
        If `None`, after normalization, each cell has a total count equal
        to the median of the *counts_per_cell* before normalization.
    counts_per_cell : `np.array`, optional (default: `None`)
        Precomputed counts per cell.
    key_n_counts : `str`, optional (default: `'n_counts'`)
        Name of the field in `adata.obs` where the total counts per cell are
        stored.
    max_proportion_per_cell : `int` (default: `None`)
        Exclude genes counts that account for more than a specific proportion of cell size, e.g. 0.05.
    use_initial_size : `bool` (default: `True`)
        Whether to use initial cell sizes oder actual cell sizes.
    layers : `str` or `list` (default: `{'spliced', 'unspliced'}`)
        Keys for layers to be also considered for normalization.
    copy : `bool`, optional (default: `False`)
        If an :class:`~anndata.AnnData` is passed, determines whether a copy
        is returned.

    Returns
    -------
    Returns or updates `adata` with normalized version of the original `adata.X`, depending on `copy`.
    """
    adata = data.copy() if copy else data
    layers = adata.layers.keys() if layers is 'all' else [layers] if isinstance(layers, str) \
        else [layer for layer in layers if layer in adata.layers.keys()]
    layers = ['X'] + layers
    modified_layers = []

    for layer in layers:
        X = adata.X if layer is 'X' else adata.layers[layer]
        if not_yet_normalized(X) or enforce:
            counts = counts_per_cell if counts_per_cell is not None \
                else get_initial_size(adata, layer) if use_initial_size else get_size(adata, layer)
            if max_proportion_per_cell is not None and (0 < max_proportion_per_cell < 1):
                counts = counts_per_cell_quantile(X, max_proportion_per_cell, counts)
            # equivalent to scanpy.pp.normalize_per_cell(X, counts_per_cell_after, counts)
            counts_after = np.median(counts) if counts_per_cell_after is None else counts_per_cell_after
            counts /= counts_after + (counts_after == 0)
            counts += counts == 0  # to avoid division by zero
            if issparse(X):
                sparsefuncs.inplace_row_scale(X, 1 / counts)
            else:
                X /= np.array(counts[:, None])
            modified_layers.append(layer)

    adata.obs['n_counts' if key_n_counts is None else key_n_counts] = get_size(adata)
    if len(modified_layers) > 0:
        logg.info('Normalized count data:', ', '.join(modified_layers) + '.')

    return adata if copy else None


def log1p(data, copy=False):
    """Logarithmize the data matrix.
    Computes :math:`X = \\log(X + 1)`, where :math:`log` denotes the natural logarithm.
    Parameters
    ----------
    data: :class:`~anndata.AnnData`
        Annotated data matrix.
    copy: `bool` (default: `False`)
        Return a copy of `adata` instead of updating it.
    Returns
    -------
    Returns or updates `adata` depending on `copy`.
    """
    adata = data.copy() if copy else data
    X = (adata.X.data if issparse(adata.X) else adata.X) if isinstance(adata, AnnData) else adata
    np.log1p(X, out=X)
    return adata if copy else None


def filter_and_normalize(data, min_counts=None, min_counts_u=None, min_cells=None, min_cells_u=None, n_top_genes=None,
                         flavor='seurat', log=True, copy=False):
    """Filtering, normalization and log transform

    Expects non-logarithmized data. If using logarithmized data, pass `log=False`.

    Runs the following steps

    .. code:: python

        scv.pp.filter_genes(adata)
        scv.pp.normalize_per_cell(adata)
        if n_top_genes is not None:
            scv.pp.filter_genes_dispersion(adata)
        if log:
            scv.pp.log1p(adata)


    Arguments
    ---------
    data: :class:`~anndata.AnnData`
        Annotated data matrix.
    min_counts: `int` (default: `None`)
        Minimum number of counts required for a gene to pass filtering (spliced).
    min_counts_u: `int` (default: `None`)
        Minimum number of counts required for a gene to pass filtering (unspliced).
    min_cells: `int` (default: `None`)
        Minimum number of cells expressed required for a gene to pass filtering (spliced).
    min_cells_u: `int` (default: `None`)
        Minimum number of cells expressed required for a gene to pass filtering (unspliced).
    n_top_genes: `int` (default: `None`)
        Number of genes to keep.
    flavor: {'seurat', 'cell_ranger', 'svr'}, optional (default: 'seurat')
        Choose the flavor for computing normalized dispersion. If choosing 'seurat', this expects non-logarithmized data.
    log: `bool` (default: `True`)
        Take logarithm.
    copy: `bool` (default: `False`)
        Return a copy of `adata` instead of updating it.

    Returns
    -------
    Returns or updates `adata` depending on `copy`.
    """
    adata = data.copy() if copy else data

    if 'spliced' not in adata.layers.keys() or 'unspliced' not in adata.layers.keys():
        raise ValueError('Could not find spliced / unspliced counts.')

    filter_genes(adata, min_counts=min_counts, min_counts_u=min_counts_u, min_cells=min_cells, min_cells_u=min_cells_u)
    normalize_per_cell(adata)
    if n_top_genes is not None:
        filter_genes_dispersion(adata, n_top_genes=n_top_genes, flavor=flavor)

    log_advised = np.allclose(adata.X[:10].sum(), adata.layers['spliced'][:10].sum())
    if log and log_advised:
        log1p(adata)

    logg.info('Logarithmized X.' if log and log_advised else
              'Did not modify X as it looks preprocessed already.' if log else
              'Consider logarithmizing X with `scv.pp.log1p` for better results.' if log_advised else '')

    return adata if copy else None


def recipe_velocity(adata, min_counts=3, min_counts_u=3, n_top_genes=None, n_pcs=30, n_neighbors=30, log=True, copy=False):
    """Runs pp.filter_and_normalize() and pp.moments()
    """
    from .moments import moments
    filter_and_normalize(adata, min_counts=min_counts, min_counts_u=min_counts_u, n_top_genes=n_top_genes, log=log)
    moments(adata, n_neighbors=n_neighbors, n_pcs=n_pcs)
    return adata if copy else None



import numpy as np
from scipy.sparse import issparse, spdiags, linalg

from .utils import groups_to_bool, scale, strings_to_categoricals
from ..preprocessing.moments import get_connectivities

try: from scanpy.tools.dpt import DPT
except ImportError:
    try: from scanpy.tools._dpt import DPT
    except ImportError: pass


def principal_curve(data, basis='pca', n_comps=4, clusters_list=None, copy=False):
    """Computes the principal curve
    
    Arguments
    ---------
    data: :class:`~anndata.AnnData`
        Annotated data matrix.
    basis: `str` (default: `'pca'`)
        Basis to use for computing the principal curve.
    n_comps: `int` (default: 4)
        Number of pricipal components to be used.
    copy: `bool`, (default: `False`)
        Return a copy instead of writing to adata.
    
    Returns
    -------
    Returns or updates `adata` with the attributes
    principal_curve: `.uns`
        dictionary containing `projections`, `ixsort` and `arclength`
    """
    adata = data.copy() if copy else data
    import rpy2.robjects as robjects
    from rpy2.robjects.packages import importr

    if clusters_list is not None:
        cell_subset = np.array([label in clusters_list for label in adata.obs['clusters']])
        X_emb = adata[cell_subset].obsm['X_' + basis][:, :n_comps]
    else:
        cell_subset = None
        X_emb = adata.obsm['X_' + basis][:, :n_comps]

    n_obs, n_dim = X_emb.shape

    # convert array to R matrix
    xvec = robjects.FloatVector(X_emb.T.reshape((X_emb.size)))
    X_R = robjects.r.matrix(xvec, nrow=n_obs, ncol=n_dim)

    fit = importr("princurve").principal_curve(X_R)

    adata.uns['principal_curve'] = dict()
    adata.uns['principal_curve']['ixsort'] = ixsort = np.array(fit[1])-1
    adata.uns['principal_curve']['projections'] = np.array(fit[0])[ixsort]
    adata.uns['principal_curve']['arclength'] = np.array(fit[2])
    adata.uns['principal_curve']['cell_subset'] = cell_subset

    return adata if copy else None


def velocity_map(adata=None, T=None, n_dcs=10, return_model=False):
    vpt = VPT(adata, n_dcs=n_dcs)
    if T is None:
        T = adata.uns['velocity_graph'] - adata.uns['velocity_graph_neg']
        vpt._connectivities = T + T.T
    vpt.compute_transitions()
    vpt.compute_eigen(n_dcs)
    adata.obsm['X_vmap'] = vpt.eigen_basis
    return vpt if return_model else None


class VPT(DPT):
    def set_iroot(self, root=None):
        if isinstance(root, str) and root in self._adata.obs.keys() and self._adata.obs[root].max() != 0:
            self.iroot = scale(get_connectivities(self._adata).dot(self._adata.obs[root])).argmax()
        elif isinstance(root, str) and root in self._adata.obs_names:
            self.iroot = np.where(self._adata.obs_names == root)[0][0]
        elif isinstance(root, (int, np.integer)) and root < self._adata.n_obs:
            self.iroot = root
        else:
            self.iroot = None

    def compute_transitions(self, density_normalize=True):
        T = self._connectivities
        if density_normalize:
            q = np.asarray(T.sum(axis=0))
            Q = spdiags(1.0 / q, 0, T.shape[0], T.shape[0]) if issparse(T) else np.diag(1.0 / q)
            K = Q.dot(T).dot(Q)
        else:
            K = T
        z = np.sqrt(np.asarray(K.sum(axis=0)))
        Z = spdiags(1.0 / z, 0, K.shape[0], K.shape[0]) if issparse(K) else np.diag(1.0 / z)
        self._transitions_sym = Z.dot(K).dot(Z)

    def compute_eigen(self, n_comps=10, sym=None, sort='decrease'):
        if self._transitions_sym is None:
            raise ValueError('Run `.compute_transitions` first.')
        n_comps = min(self._transitions_sym.shape[0] - 1, n_comps)
        evals, evecs = linalg.eigsh(self._transitions_sym, k=n_comps, which='LM')
        self._eigen_values = evals[::-1]
        self._eigen_basis = evecs[:, ::-1]

    def compute_pseudotime(self, inverse=False):
        if self.iroot is not None:
            self._set_pseudotime()
            self.pseudotime = 1 - self.pseudotime if inverse else self.pseudotime
            self.pseudotime[~np.isfinite(self.pseudotime)] = np.nan
        else:
            self.pseudotime = np.empty(self._adata.n_obs)
            self.pseudotime[:] = np.nan


def velocity_pseudotime(adata, groupby=None, groups=None, root=None, end=None, n_dcs=10, n_branchings=0,
                        min_group_size=0.01, allow_kendall_tau_shift=True, use_velocity_field=True,
                        save_diffmap=False, return_model=False):
    strings_to_categoricals(adata)
    root = 'root_cells' if root is None and 'root_cells' in adata.obs.keys() else root
    end = 'end_points' if end is None and 'end_points' in adata.obs.keys() else end
    groupby = 'cell_fate' if groupby is None and 'cell_fate' in adata.obs.keys() else groupby
    categories = adata.obs[groupby].cat.categories if groupby is not None and groups is None else [None]
    for cat in categories:
        groups = cat if cat is not None else groups
        cell_subset = groups_to_bool(adata, groups=groups, groupby=groupby)
        data = adata.copy() if cell_subset is None else adata[cell_subset].copy()
        vpt = VPT(data, n_dcs=n_dcs, min_group_size=min_group_size,
                  n_branchings=n_branchings, allow_kendall_tau_shift=allow_kendall_tau_shift)

        if use_velocity_field:
            T = data.uns['velocity_graph'] - data.uns['velocity_graph_neg']
            vpt._connectivities = T + T.T

        vpt.compute_transitions()
        vpt.compute_eigen(n_comps=n_dcs)

        vpt.set_iroot(root)
        vpt.compute_pseudotime()
        dpt_root = vpt.pseudotime

        vpt.set_iroot(end)
        vpt.compute_pseudotime(inverse=True)
        dpt_end = vpt.pseudotime

        # merge dpt_root and inverse dpt_end together
        vpt.pseudotime = np.nan_to_num(dpt_root) + np.nan_to_num(dpt_end)
        vpt.pseudotime[np.isfinite(dpt_root) & np.isfinite(dpt_end)] /= 2
        vpt.pseudotime = scale(vpt.pseudotime)
        vpt.pseudotime[np.isnan(dpt_root) & np.isnan(dpt_end)] = np.nan

        if n_branchings > 0: vpt.branchings_segments()
        else: vpt.indices = vpt.pseudotime.argsort()

        if 'velocity_pseudotime' not in adata.obs.keys():
            pseudotime = np.empty(adata.n_obs)
            pseudotime[:] = np.nan
        else:
            pseudotime = adata.obs['velocity_pseudotime'].copy()
        pseudotime[cell_subset] = vpt.pseudotime
        adata.obs['velocity_pseudotime'] = pseudotime

        if save_diffmap:
            diffmap = np.empty(shape=(adata.n_obs, n_dcs))
            diffmap[:] = np.nan
            diffmap[cell_subset] = vpt.eigen_basis
            adata.obsm['X_diffmap_' + groups] = diffmap

    return vpt if return_model else None



from .. import settings
from .. import logging as logg

from scanpy.api import Neighbors
from scanpy.api.pp import pca
from scipy.sparse import issparse
import numpy as np


def neighbors(adata, n_neighbors=30, n_pcs=30, use_rep=None, knn=True, random_state=0, method='umap',
              metric='euclidean', metric_kwds={}, copy=False):
    """
    Compute a neighborhood graph of observations [McInnes18]_.
    The neighbor search efficiency of this heavily relies on UMAP [McInnes18]_,
    which also provides a method for estimating connectivities of data points -
    the connectivity of the manifold (`method=='umap'`). If `method=='diffmap'`,
    connectivities are computed according to [Coifman05]_, in the adaption of
    [Haghverdi16]_.
    Parameters
    ----------
    adata
        Annotated data matrix.
    n_neighbors
        The size of local neighborhood (in terms of number of neighboring data
        points) used for manifold approximation. Larger values result in more
        global views of the manifold, while smaller values result in more local
        data being preserved. In general values should be in the range 2 to 100.
        If `knn` is `True`, number of nearest neighbors to be searched. If `knn`
        is `False`, a Gaussian kernel width is set to the distance of the
        `n_neighbors` neighbor.
    n_pcs : `int` or `None` (default: None)
        Use this many PCs. If n_pcs==0 use .X if use_rep is None.

    use_rep : `None`, `'X'` or any key for `.obsm` (default: None)
        Use the indicated representation. If `None`, the representation is chosen automatically:
        for .n_vars < 50, .X is used, otherwise ‘X_pca’ is used.
    knn
        If `True`, use a hard threshold to restrict the number of neighbors to
        `n_neighbors`, that is, consider a knn graph. Otherwise, use a Gaussian
        Kernel to assign low weights to neighbors more distant than the
        `n_neighbors` nearest neighbor.
    random_state
        A numpy random seed.
    method : {{'umap', 'gauss', `sklearn`, `None`}}  (default: `'umap'`)
        Use 'umap' [McInnes18]_ or 'gauss' (Gauss kernel following [Coifman05]_
        with adaptive width [Haghverdi16]_) for computing connectivities.
    metric
        A known metric’s name or a callable that returns a distance.
    metric_kwds
        Options for the metric.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Depending on `copy`, updates or returns `adata` with the following:
    connectivities : sparse matrix (`.uns['neighbors']`, dtype `float32`)
        Weighted adjacency matrix of the neighborhood graph of data
        points. Weights should be interpreted as connectivities.
    distances : sparse matrix (`.uns['neighbors']`, dtype `float32`)
        Instead of decaying weights, this stores distances for each pair of
        neighbors.
    """
    logg.info('computing neighbors', r=True)
    adata = adata.copy() if copy else adata
    if adata.isview: adata._init_as_actual(adata.copy())

    if (use_rep is None or use_rep is 'X_pca') \
            and ('X_pca' not in adata.obsm.keys() or n_pcs > adata.obsm['X_pca'].shape[1]):
        pca(adata, n_comps=n_pcs, svd_solver='arpack')

    adata.uns['neighbors'] = {}
    adata.uns['neighbors']['params'] = {'n_neighbors': n_neighbors, 'method': method}

    if method is 'sklearn':
        from sklearn.neighbors import NearestNeighbors
        neighbors = NearestNeighbors(n_neighbors=n_neighbors)
        neighbors.fit(adata.obsm['X_pca'] if use_rep is None else adata.obsm[use_rep])
        adata.uns['neighbors']['distances'] = neighbors.kneighbors_graph(mode='distance')
        adata.uns['neighbors']['connectivities'] = neighbors.kneighbors_graph(mode='connectivity')

    else:
        neighbors = Neighbors(adata)
        neighbors.compute_neighbors(n_neighbors=n_neighbors, knn=knn, n_pcs=n_pcs, use_rep=use_rep, method=method,
                                    metric=metric, metric_kwds=metric_kwds, random_state=random_state, write_knn_indices=True)
        adata.uns['neighbors']['distances'] = neighbors.distances
        adata.uns['neighbors']['connectivities'] = neighbors.connectivities
        adata.uns['neighbors']['indices'] = neighbors.knn_indices

    logg.info('    finished', time=True, end=' ' if settings.verbosity > 2 else '\n')
    logg.hint(
        'added to `.uns[\'neighbors\']`\n'
        '    \'distances\', weighted adjacency matrix\n'
        '    \'connectivities\', weighted adjacency matrix')
    return adata if copy else None


def select_distances(dist, n_neighbors=None):
    D = dist.copy()
    n_counts = (D > 0).sum(1).A1 if issparse(D) else (D > 0).sum(1)
    n_neighbors = n_counts.min() if n_neighbors is None else min(n_counts.min(), n_neighbors)
    rows = np.where(n_counts > n_neighbors)[0]
    cumsum_neighs = np.insert(n_counts.cumsum(), 0, 0)
    dat = D.data

    for row in rows:
        n0, n1 = cumsum_neighs[row], cumsum_neighs[row + 1]
        rm_idx = n0 + dat[n0:n1].argsort()[n_neighbors:]
        dat[rm_idx] = 0
    D.eliminate_zeros()
    return D


def select_connectivities(connectivities, n_neighbors=None):
    C = connectivities.copy()
    n_counts = (C > 0).sum(1).A1 if issparse(C) else (C > 0).sum(1)
    n_neighbors = n_counts.min() if n_neighbors is None else min(n_counts.min(), n_neighbors)
    rows = np.where(n_counts > n_neighbors)[0]
    cumsum_neighs = np.insert(n_counts.cumsum(), 0, 0)
    dat = C.data

    for row in rows:
        n0, n1 = cumsum_neighs[row], cumsum_neighs[row + 1]
        rm_idx = n0 + dat[n0:n1].argsort()[::-1][n_neighbors:]
        dat[rm_idx] = 0
    C.eliminate_zeros()
    return C


def neighbors_to_be_recomputed(adata, n_neighbors=None):
    # check if neighbors graph is disrupted
    n_neighs = (adata.uns['neighbors']['distances'] > 0).sum(1)
    result = n_neighs.max() - n_neighs.min() >= 2
    # check if neighbors graph has sufficient number of neighbors
    if n_neighbors is not None:
        result = result or n_neighbors > adata.uns['neighbors']['params']['n_neighbors']
    return result


def get_connectivities(adata, mode='connectivities', n_neighbors=None, recurse_neighbors=False):
    C = adata.uns['neighbors'][mode]
    if n_neighbors is not None and n_neighbors < adata.uns['neighbors']['params']['n_neighbors']:
        C = select_connectivities(C, n_neighbors) if mode == 'connectivities' else select_distances(C, n_neighbors)
    connectivities = C > 0
    connectivities.setdiag(1)
    if recurse_neighbors:
        connectivities += connectivities.dot(connectivities * .5)
        connectivities.data = np.clip(connectivities.data, 0, 1)
    connectivities = connectivities.multiply(1. / connectivities.sum(1))
    return connectivities.tocsr().astype(np.float32)


from .scatter import scatter
from .velocity import velocity
from .velocity_embedding import velocity_embedding
from .velocity_embedding_grid import velocity_embedding_grid
from .velocity_embedding_stream import velocity_embedding_stream
from .velocity_graph import velocity_graph
from .heatmap import heatmap
from .utils import hist, plot
from .simulation import simulation
from scanpy.api.pl import paga, paga_compare, rank_genes_groups



from .utils import sum_obs, prod_sum_obs, make_dense
from scipy.optimize import minimize
from scipy.sparse import csr_matrix, issparse
import numpy as np
import warnings


def get_weight(x, y, perc):
    if issparse(x): x = x.A
    if issparse(y): y = y.A
    xy_norm = x / np.clip(x.max(0), 1e-3, None) + y / np.clip(y.max(0), 1e-3, None)
    if isinstance(perc, int):
        weights = xy_norm >= np.percentile(xy_norm, perc, axis=0)
    else:
        lb, ub = np.percentile(xy_norm, perc, axis=0)
        weights = (xy_norm <= lb) | (xy_norm >= ub)
    return weights


def leastsq_NxN(x, y, fit_offset=False, perc=None):
    """Solution to least squares: gamma = cov(X,Y) / var(X)
    """
    if perc is not None:
        if not fit_offset and isinstance(perc, (list, tuple)): perc = perc[1]
        weights = csr_matrix(get_weight(x, y, perc)).astype(bool)
        x, y = weights.multiply(x).tocsr(), weights.multiply(y).tocsr()
    else:
        weights = None

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        xx_ = prod_sum_obs(x, x)
        xy_ = prod_sum_obs(x, y)

        if fit_offset:
            n_obs = x.shape[0] if weights is None else sum_obs(weights)
            x_ = sum_obs(x) / n_obs
            y_ = sum_obs(y) / n_obs
            gamma = (xy_ / n_obs - x_ * y_) / (xx_ / n_obs - x_ ** 2)
            offset = y_ - gamma * x_

            # fix negative offsets:
            idx = offset < 0
            gamma[idx] = xy_[idx] / xx_[idx]
            offset = np.clip(offset, 0, None)
        else:
            gamma = xy_ / xx_
            offset = np.zeros(x.shape[1])
    offset[np.isnan(offset)], gamma[np.isnan(gamma)] = 0, 0
    return offset, gamma


def optimize_NxN(x, y, fit_offset=False, perc=None):
    """Just to compare with closed-form solution
    """
    if perc is not None:
        if not fit_offset and isinstance(perc, (list, tuple)): perc = perc[1]
        weights = get_weight(x, y, perc).astype(bool)
        if issparse(weights): weights = weights.A
    else:
        weights = None

    x, y = x.astype(np.float64), y.astype(np.float64)

    n_vars = x.shape[1]
    offset, gamma = np.zeros(n_vars), np.zeros(n_vars)

    for i in range(n_vars):
        xi = x[:, i] if weights is None else x[:, i][weights[:, i]]
        yi = y[:, i] if weights is None else y[:, i][weights[:, i]]

        if fit_offset:
            offset[i], gamma[i] = minimize(lambda m: np.sum((-yi + xi * m[1] + m[0])**2), method="L-BFGS-B",
                                           x0=(0, 0.1), bounds=[(0, None), (None, None)]).x
        else:
            gamma[i] = minimize(lambda m: np.sum((-yi + xi * m) ** 2), x0=0.1, method="L-BFGS-B").x
    offset[np.isnan(offset)], gamma[np.isnan(gamma)] = 0, 0
    return offset, gamma


def leastsq_generalized(x, y, x2, y2, res_std=None, res2_std=None, fit_offset=False, fit_offset2=False, perc=None):
    """Solution to the 2-dim generalized least squares: gamma = inv(X'QX)X'QY
    """
    if perc is not None:
        if not fit_offset and isinstance(perc, (list, tuple)): perc = perc[1]
        weights = csr_matrix(get_weight(x, y, perc)).astype(bool)
        x, y = weights.multiply(x).tocsr(), weights.multiply(y).tocsr()

    n_obs, n_var = x.shape
    offset, offset_ss = np.zeros(n_var, dtype="float32"), np.zeros(n_var, dtype="float32")
    gamma = np.ones(n_var, dtype="float32")

    if (res_std is None) or (res2_std is None): res_std, res2_std = np.ones(n_var), np.ones(n_var)
    ones, zeros = np.ones(n_obs), np.zeros(n_obs)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        x, y = np.vstack((make_dense(x)/res_std, x2/res2_std)), np.vstack((make_dense(y)/res_std, y2/res2_std))

    if fit_offset and fit_offset2:
        for i in range(n_var):
            A = np.c_[np.vstack((np.c_[ones/res_std[i], zeros], np.c_[zeros, ones/res2_std[i]])), x[:, i]]
            offset[i], offset_ss[i], gamma[i] = np.linalg.pinv(A.T.dot(A)).dot(A.T.dot(y[:, i]))
    elif fit_offset:
        for i in range(n_var):
            A = np.c_[np.hstack((ones/res_std[i], zeros)), x[:, i]]
            offset[i], gamma[i] = np.linalg.pinv(A.T.dot(A)).dot(A.T.dot(y[:, i]))
    elif fit_offset2:
        for i in range(n_var):
            A = np.c_[np.hstack((zeros, ones/res2_std[i])), x[:, i]]
            offset_ss[i], gamma[i] = np.linalg.pinv(A.T.dot(A)).dot(A.T.dot(y[:, i]))
    else:
        for i in range(n_var):
            A = np.c_[x[:, i]]
            gamma[i] = np.linalg.pinv(A.T.dot(A)).dot(A.T.dot(y[:, i]))

    offset[np.isnan(offset)] = 0
    offset_ss[np.isnan(offset_ss)] = 0
    gamma[np.isnan(gamma)] = 0

    return offset, offset_ss, gamma


def maximum_likelihood(Ms, Mu, Mus, Mss, fit_offset=False, fit_offset2=False):
    """Maximizing the log likelihood using weights according to empirical bayes
    """
    n_obs, n_var = Ms.shape
    offset, offset_ss = np.zeros(n_var, dtype="float32"), np.zeros(n_var, dtype="float32")
    gamma = np.ones(n_var, dtype="float32")

    def sse(A, data, b):
        sigma = (A.dot(data) - b).std(1)
        return np.log(sigma).sum()  # np.log(np.sqrt(2*np.pi) * sigma).sum() + (.5 * (res/sigma[:, None])**2).sum()

    if fit_offset and fit_offset2:
        for i in range(n_var):
            data = np.vstack((Mu[:, i], Ms[:, i], Mus[:, i], Mss[:, i]))
            offset[i], offset_ss[i], gamma[i] = \
                minimize(lambda m: sse(np.array([[1, -m[2], 0, 0], [1, m[2], 2, -2 * m[2]]]),
                                       data, b=np.array(m[0], m[1])), x0=(1e-4, 1e-4, 1), method="L-BFGS-B").x
    elif fit_offset:
        for i in range(n_var):
            data = np.vstack((Mu[:, i], Ms[:, i], Mus[:, i], Mss[:, i]))
            offset[i], gamma[i] = \
                minimize(lambda m: sse(np.array([[1, -m[1], 0, 0], [1, m[1], 2, -2 * m[1]]]),
                                       data, b=np.array(m[0], 0)), x0=(1e-4, 1), method="L-BFGS-B").x
    elif fit_offset2:
        for i in range(n_var):
            data = np.vstack((Mu[:, i], Ms[:, i], Mus[:, i], Mss[:, i]))
            offset_ss[i], gamma[i] = \
                minimize(lambda m: sse(np.array([[1, -m[1], 0, 0], [1, m[1], 2, -2 * m[1]]]),
                                       data, b=np.array(0, m[0])), x0=(1e-4, 1), method="L-BFGS-B").x
    else:
        for i in range(n_var):
            data = np.vstack((Mu[:, i], Ms[:, i], Mus[:, i], Mss[:, i]))
            gamma[i] = \
                minimize(lambda m: sse(np.array([[1, -m, 0, 0], [1, m, 2, -2 * m]]), data, b=0),
                         x0=gamma[i], method="L-BFGS-B").x
    return offset, offset_ss, gamma



from scipy.sparse import csr_matrix, issparse
import pandas as pd
import numpy as np
import warnings


def mean(x, axis=0):
    return x.mean(axis).A1 if issparse(x) else x.mean(axis)


def make_dense(X):
    return X.A if issparse(X) and X.ndim==2 else X.A1 if issparse(X) else X


def sum_obs(A):
    """summation over axis 0 (obs) equivalent to np.sum(A, 0)
    """
    return A.sum(0).A1 if issparse(A) else np.einsum('ij -> j', A)


def prod_sum_obs(A, B):
    """dot product and sum over axis 0 (obs) equivalent to np.sum(A * B, 0)
    """
    return A.multiply(B).sum(0).A1 if issparse(A) else np.einsum('ij, ij -> j', A, B)


def prod_sum_var(A, B):
    """dot product and sum over axis 1 (var) equivalent to np.sum(A * B, 1)
    """
    return A.multiply(B).sum(1).A1 if issparse(A) else np.einsum('ij, ij -> i', A, B)


def norm(A):
    """computes the L2-norm along axis 1 (e.g. genes or embedding dimensions) equivalent to np.linalg.norm(A, axis=1)
    """
    return np.sqrt(A.multiply(A).sum(1).A1) if issparse(A) else np.sqrt(np.einsum('ij, ij -> i', A, A))


def vector_norm(x):
    """computes the L2-norm along axis 1 (e.g. genes or embedding dimensions) equivalent to np.linalg.norm(A, axis=1)
    """
    return np.sqrt(np.einsum('i, i -> ', x, x))


def R_squared(residual, total):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        r2 = np.ones(residual.shape[1]) - prod_sum_obs(residual, residual) / prod_sum_obs(total, total)
    r2[np.isnan(r2)] = 0
    return r2


def cosine_correlation(dX, Vi):
    dX -= dX.mean(-1)[:, None]
    Vi_norm = vector_norm(Vi)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        result = np.zeros(dX.shape[0]) if Vi_norm == 0 else np.einsum('ij, j', dX, Vi) / (norm(dX) * Vi_norm)[None, :]
    return result


def normalize(X):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        X = X.multiply(csr_matrix(1. / np.abs(X).sum(1))) if issparse(X) else X / X.sum(1)
    return X


def scale(X, min=0, max=1):
    idx = np.isfinite(X)
    if any(idx):
        X = X - X[idx].min() + min
        xmax = X[idx].max()
        X = X / xmax * max if xmax != 0 else X * max
    return X


def get_indices(dist, n_neighbors=None):
    D = dist.copy()
    n_counts = (D > 0).sum(1).A1 if issparse(D) else (D > 0).sum(1)
    n_neighbors = n_counts.min() if n_neighbors is None else min(n_counts.min(), n_neighbors)
    rows = np.where(n_counts > n_neighbors)[0]
    cumsum_neighs = np.insert(n_counts.cumsum(), 0, 0)
    dat = D.data

    for row in rows:
        n0, n1 = cumsum_neighs[row], cumsum_neighs[row + 1]
        rm_idx = n0 + dat[n0:n1].argsort()[n_neighbors:]
        dat[rm_idx] = 0
    D.eliminate_zeros()
    indices = D.indices.reshape((-1, n_neighbors))
    return indices, D


def get_iterative_indices(indices, index, n_recurse_neighbors=2, max_neighs=None):
    def iterate_indices(indices, index, n_recurse_neighbors):
        return indices[iterate_indices(indices, index, n_recurse_neighbors - 1)] \
            if n_recurse_neighbors > 1 else indices[index]
    indices = np.unique(iterate_indices(indices, index, n_recurse_neighbors))
    if max_neighs is not None and len(indices) > max_neighs:
        indices = np.random.choice(indices, max_neighs, replace=False)
    return indices


def groups_to_bool(adata, groups, groupby=None):
    groups = [groups] if isinstance(groups, str) else groups
    if isinstance(groups, (list, tuple, np.ndarray, np.record)):
        groupby = groupby if groupby in adata.obs.keys() else 'clusters' if 'clusters' in adata.obs.keys() \
            else 'louvain' if 'louvain' in adata.obs.keys() else None
        if groupby is not None:
            groups = np.array([key in groups for key in adata.obs[groupby]])
        else: raise ValueError('groupby attribute not valid.')
    return groups


def most_common_in_list(lst):
    lst = list(lst)
    return max(set(lst), key=lst.count)


def randomized_velocity(adata, vkey='velocity', add_key='velocity_random'):
    V_rnd = adata.layers[vkey].copy()
    for i in range(V_rnd.shape[1]):
        np.random.shuffle(V_rnd[:, i])
        V_rnd[:, i] = V_rnd[:, i] * np.random.choice(np.array([+1, -1]), size=V_rnd.shape[0])
    adata.layers[add_key] = V_rnd

    from .velocity_graph import velocity_graph
    from .velocity_embedding import velocity_embedding

    velocity_graph(adata, vkey=add_key)
    velocity_embedding(adata, vkey=add_key, autoscale=False)


def extract_int_from_str(array):
    def str_to_int(item):
        num = "".join(filter(str.isdigit, item))
        num = int(num) if len(num) > 0 else -1
        return num
    if isinstance(array, str): nums = str_to_int(array)
    elif len(array) > 1 and isinstance(array[0], str):
        nums = []
        for item in array: nums.append(str_to_int(item))
    else: nums = array
    nums = pd.Categorical(nums) if array.dtype == 'category' else np.array(nums)
    return nums


def strings_to_categoricals(adata):
    """Transform string annotations to categoricals.
    """
    from pandas import Categorical

    def is_valid_dtype(values):
        from pandas.api.types import is_string_dtype, is_integer_dtype, is_bool_dtype
        return is_string_dtype(values) or is_integer_dtype(values) or is_bool_dtype(values)

    for df in [adata.obs, adata.var]:
        string_cols = [key for key in df.columns if is_valid_dtype(df[key])]
        for key in string_cols:
            c = df[key]
            c = Categorical(c)
            if len(c.categories) < min(len(c), 100): df[key] = c


def merge_groups(adata, key, map_groups, key_added=None, map_colors=None):
    strings_to_categoricals(adata)
    if len(map_groups) != len(adata.obs[key].cat.categories):
        map_coarse = {}
        for c in adata.obs[key].cat.categories:
            for group in map_groups:
                if any(cluster == c for cluster in map_groups[group]): map_coarse[c] = group
            if c not in map_coarse: map_coarse[c] = c
        map_groups = map_coarse

    if key_added is None:
        key_added = key + '_coarse'

    from pandas.api.types import CategoricalDtype
    adata.obs[key_added] = adata.obs[key].map(map_groups).astype(CategoricalDtype())
    old_categories = adata.obs[key].cat.categories
    new_categories = adata.obs[key_added].cat.categories

    # map_colors is passed
    if map_colors is not None:
        old_colors = None
        if key + '_colors' in adata.uns:
            old_colors = adata.uns[key + '_colors']
        new_colors = []
        for group in adata.obs[key_added].cat.categories:
            if group in map_colors:
                new_colors.append(map_colors[group])
            elif group in old_categories and old_colors is not None:
                new_colors.append(old_colors[old_categories.get_loc(group)])
            else:
                raise ValueError('You didn\'t specify a color for {}.'.format(group))
        adata.uns[key_added + '_colors'] = new_colors

    # map_colors is not passed
    elif key + '_colors' in adata.uns:
        old_colors = adata.uns[key + '_colors']
        inverse_map_groups = {g: [] for g in new_categories}
        for old_group in old_categories:
            inverse_map_groups[map_groups[old_group]].append(old_group)
        new_colors = []
        for group in new_categories:
            # take the largest of the old groups
            old_group = adata.obs[key][adata.obs[key].isin(
                inverse_map_groups[group])].value_counts().index[0]
            new_colors.append(old_colors[old_categories.get_loc(old_group)])
        adata.uns[key_added + '_colors'] = new_colors


def cutoff_small_velocities(adata, vkey='velocity', key_added='velocity_cut', frac_of_max=.5, use_raw=False):
    x = adata.layers['spliced'] if use_raw else adata.layers['Ms']
    y = adata.layers['unspliced'] if use_raw else adata.layers['Mu']

    x_max = x.max(0).A[0] if issparse(x) else x.max(0)
    y_max = y.max(0).A[0] if issparse(y) else y.max(0)

    xy_norm = x / np.clip(x_max, 1e-3, None) + y / np.clip(y_max, 1e-3, None)
    W = xy_norm >= np.percentile(xy_norm, 98, axis=0) * frac_of_max

    adata.layers[key_added] = csr_matrix(W).multiply(adata.layers[vkey]).tocsr()

    from .velocity_graph import velocity_graph
    from .velocity_embedding import velocity_embedding

    velocity_graph(adata, vkey=key_added, approx=True)
    velocity_embedding(adata, vkey=key_added)




from .. import settings
from .utils import is_categorical, update_axes, set_label, set_title, default_basis, default_color, default_color_map, \
    default_size, interpret_colorkey, get_components, set_colorbar, savefig_or_show, make_unique_list, show_linear_fit
from .docs import doc_scatter, doc_params

from matplotlib import rcParams
import matplotlib.pyplot as pl
import numpy as np
from scipy.sparse import issparse


@doc_params(scatter=doc_scatter)
def scatter(adata, x=None, y=None, basis=None, vkey=None, color=None, use_raw=None, layer=None, color_map=None,
            colorbar=True, palette=None, size=None, alpha=None, perc=None, sort_order=True, groups=None,
            components=None, projection='2d', legend_loc='none', legend_fontsize=None, legend_fontweight=None,
            right_margin=None, left_margin=None, xlabel=None, ylabel=None, title=None, fontsize=None, figsize=None,
            dpi=None, frameon=None, show=True, save=None, ax=None, zorder=None, ncols=None, **kwargs):
    """\
    Scatter plot along observations or variables axes.

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    x: `str`, `np.ndarray` or `None` (default: `None`)
        x coordinate
    y: `str`, `np.ndarray` or `None` (default: `None`)
        y coordinate
    {scatter}

    Returns
    -------
        If `show==False` a `matplotlib.Axis`
    """
    scatter_kwargs = {"use_raw": use_raw, "sort_order": sort_order, "alpha": alpha, "components": components,
                      "projection": projection, "legend_loc": legend_loc, "groups": groups, "palette": palette,
                      "legend_fontsize": legend_fontsize, "legend_fontweight": legend_fontweight,
                      "right_margin": right_margin, "left_margin": left_margin, "show": False, "save": None}

    colors, layers, bases = make_unique_list(color, allow_array=True), make_unique_list(layer), make_unique_list(basis)
    multikey = colors if len(colors) > 1 else layers if len(layers) > 1 else bases if len(bases) > 1 else None
    if multikey is not None:
        if isinstance(title, (list, tuple)): title *= int(np.ceil(len(multikey) / len(title)))
        ncols = len(multikey) if ncols is None else min(len(multikey), ncols)
        nrows = int(np.ceil(len(multikey) / ncols))
        figsize = rcParams['figure.figsize'] if figsize is None else figsize
        for i, gs in enumerate(
                pl.GridSpec(nrows, ncols, pl.figure(None, (figsize[0] * ncols, figsize[1] * nrows), dpi=dpi))):
            if i < len(multikey):
                scatter(adata, x=x, y=y, size=size, xlabel=xlabel, ylabel=ylabel, color_map=color_map,
                        colorbar=colorbar, perc=perc, frameon=frameon, ax=pl.subplot(gs), zorder=zorder,
                        color=colors[i] if len(colors) > 1 else color,
                        layer=layers[i] if len(layers) > 1 else layer,
                        basis=bases[i] if len(bases) > 1 else basis,
                        title=title[i] if isinstance(title, (list, tuple)) else title, **scatter_kwargs, **kwargs)
        savefig_or_show('' if basis is None else basis, dpi=dpi, save=save, show=show)
        if not show: return ax

    else:
        color, layer, basis = colors[0], layers[0], bases[0]
        color = default_color(adata) if color is None else color
        color_map = default_color_map(adata, color) if color_map is None else color_map
        frameon = frameon if frameon is not None else settings._frameon

        is_embedding = ((x is None) | (y is None)) and basis not in adata.var_names
        basis = default_basis(adata) if basis is None and is_embedding else basis
        size = default_size(adata) if size is None else size

        if projection == '3d':
            from mpl_toolkits.mplot3d import Axes3D
            ax = pl.figure(None, figsize, dpi=dpi).gca(projection=projection) if ax is None else ax
        else:
            ax = pl.figure(None, figsize, dpi=dpi).gca() if ax is None else ax

        if is_categorical(adata, color) and is_embedding:
            from scanpy.api.pl import scatter as scatter_
            ax = scatter_(adata, basis=basis, color=color, color_map=color_map, size=size, frameon=frameon, ax=ax,
                          title=title, **scatter_kwargs, **kwargs)

        else:
            if basis in adata.var_names:
                x = adata[:, basis].layers['spliced'] if use_raw else adata[:, basis].layers['Ms']
                y = adata[:, basis].layers['unspliced'] if use_raw else adata[:, basis].layers['Mu']
                x, y = x.A if issparse(x) else x, y.A if issparse(y) else y
                xlabel, ylabel = 'spliced', 'unspliced'
                title = basis if title is None else title

            elif is_embedding:
                X_emb = adata.obsm['X_' + basis][:, get_components(components, basis)]
                x, y = X_emb[:, 0], X_emb[:, 1]

            elif isinstance(x, str) and isinstance(y, str) and x in adata.var_names and y in adata.var_names:
                x = adata[:, x].layers[layer] if layer in adata.layers.keys() else adata[:, x].X
                y = adata[:, y].layers[layer] if layer in adata.layers.keys() else adata[:, y].X

            c = interpret_colorkey(adata, basis, color, perc) if basis in adata.var_names and color in adata.layers.keys() \
                else interpret_colorkey(adata, color, layer, perc)

            if layer is not None and 'velocity' in layer and isinstance(color, str) and color in adata.var_names:
                ub = np.percentile(np.abs(c), 98)
                kwargs.update({"vmin": -ub, "vmax": ub})
            if layer is not None and ('spliced' in layer or 'Ms' in layer or 'Mu' in layer) \
                    and isinstance(color, str) and color in adata.var_names:
                ub = np.percentile(c, 98)
                kwargs.update({"vmax": ub})

            if groups is not None or (not isinstance(c, str) and not isinstance(c[0], str) and any(np.isnan(c))):
                zorder = 0 if zorder is None else zorder
                ax = scatter(adata, basis=basis, color='lightgrey', ax=ax, zorder=zorder, **scatter_kwargs)
                zorder += 1

            pl.scatter(x, y, c=c, cmap=color_map, s=size, alpha=alpha, edgecolors='none', marker='.', zorder=zorder, **kwargs)

            set_label(xlabel, ylabel, fontsize, basis)
            set_title(title, layer, color, fontsize)
            ax = update_axes(ax, fontsize, is_embedding, frameon)

            if basis in adata.var_names:
                show_linear_fit(adata, basis, vkey, x)

                from .simulation import show_full_dynamics
                if 'true_alpha' in adata.var.keys():
                    show_full_dynamics(adata, basis)
                if 'fit_alpha' in adata.var.keys():
                    show_full_dynamics(adata, basis, 'fit')

            if colorbar and not is_categorical(adata, color): set_colorbar(ax)

        savefig_or_show('' if basis is None else basis, dpi=dpi, save=save, show=show)
        if not show: return ax


from .utils import show_proportions, cleanup, filter_genes, filter_genes_dispersion, normalize_per_cell, log1p, \
    filter_and_normalize, recipe_velocity
from .neighbors import pca, neighbors
from .moments import moments



from .. import settings
from ..tools.transition_matrix import transition_matrix
from .utils import savefig_or_show, default_basis
from .scatter import scatter
from .docs import doc_scatter, doc_params

import warnings
import numpy as np
import matplotlib.pyplot as pl


@doc_params(scatter=doc_scatter)
def velocity_graph(adata, basis=None, vkey='velocity', which_graph='velocity', n_neighbors=10,
                   alpha=.8, perc=90, edge_width=.2, edge_color='grey', color=None, use_raw=None, layer=None,
                   color_map=None, colorbar=True, palette=None, size=None,  sort_order=True, groups=None,
                   components=None, projection='2d', legend_loc='on data', legend_fontsize=None, legend_fontweight=None,
                   right_margin=None, left_margin=None, xlabel=None, ylabel=None, title=None, fontsize=None,
                   figsize=None, dpi=None, frameon=None, show=True, save=None, ax=None):
    """\
    Plot of the velocity graph.

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    vkey: `str` or `None` (default: `None`)
        Key for annotations of observations/cells or variables/genes.
    which_graph: `'velocity'` or `'neighbors'` (default: `'velocity'`)
        Whether to show transitions from velocity graph or connectivities from neighbors graph.

    {scatter}

    Returns
    -------
        `matplotlib.Axis` if `show==False`
    """
    basis = default_basis(adata) if basis is None else basis
    title = which_graph + ' graph' if title is None else title
    scatter_kwargs = {"basis": basis, "perc": perc, "use_raw": use_raw, "sort_order": sort_order, "alpha": alpha,
                      "components": components, "projection": projection, "legend_loc": legend_loc, "groups": groups,
                      "legend_fontsize": legend_fontsize, "legend_fontweight": legend_fontweight, "palette": palette,
                      "color_map": color_map, "frameon": frameon, "title": title, "xlabel": xlabel, "ylabel": ylabel,
                      "right_margin": right_margin, "left_margin": left_margin, "colorbar": colorbar, "dpi": dpi,
                      "fontsize": fontsize, "show": False, "save": None, "figsize": figsize, }
    ax = scatter(adata, layer=layer, color=color, size=size, ax=ax, zorder=0, **scatter_kwargs)

    from networkx import Graph, draw_networkx_edges
    if which_graph == 'neighbors':
        T = adata.uns['neighbors']['connectivities']
        if perc is not None:
            threshold = np.percentile(T.data, perc)
            T.data[T.data < threshold] = 0
            T.eliminate_zeros()
    else:
        T = transition_matrix(adata, vkey=vkey, weight_indirect_neighbors=0, n_neighbors=n_neighbors, perc=perc)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        edges = draw_networkx_edges(Graph(T), adata.obsm['X_' + basis], width=edge_width, edge_color=edge_color, ax=ax)
        edges.set_zorder(-2)
        edges.set_rasterized(settings._vector_friendly)

    savefig_or_show('' if basis is None else basis, dpi=dpi, save=save, show=show)
    if not show: return ax



import numpy as np
import matplotlib.pyplot as pl
from matplotlib import rcParams


def compute_dynamics(adata, basis, key='true'):
    from ..tools.dynamical_model_utils import unspliced, spliced, vectorize
    idx = np.where(adata.var_names == basis)[0][0] if isinstance(basis, str) else basis

    alpha = adata.var[key + '_alpha'][idx]
    beta = adata.var[key + '_beta'][idx]
    gamma = adata.var[key + '_gamma'][idx]
    t_ = adata.var[key + '_t_'][idx]

    t = adata.obs[key + '_t'].values if key is 'true' else adata.layers[key + '_t'][:, idx]
    tau, alpha, u0, s0 = vectorize(np.sort(t), t_, alpha, beta, gamma)

    ut = unspliced(tau, u0, alpha, beta)
    st = spliced(tau, s0, u0, alpha, beta, gamma)

    vt = spliced(tau + np.max(tau)/100, s0, u0, alpha, beta, gamma) - st
    return alpha, ut, st, vt


def show_full_dynamics(adata, basis, key='true'):
    color = 'grey' if key is 'true' else 'purple'
    linewidth = .5 if key is 'true' else 1

    _, ut, st, _ = compute_dynamics(adata, basis, key)
    pl.plot(st, ut, color=color, linewidth=linewidth)

    idx = np.where(adata.var_names == basis)[0][0]
    beta, gamma = adata.var[key + '_beta'][idx], adata.var[key + '_gamma'][idx]
    xnew = np.linspace(0, st.max())
    pl.plot(xnew, gamma / beta * xnew, color=color, linestyle='--', linewidth=linewidth)


def simulation(adata, var_names='all', ):
    from ..tools.utils import make_dense

    idx_sorted = np.argsort(adata.obs.true_t)
    data = adata[idx_sorted]

    t = data.obs.true_t
    var_names = adata.var_names if var_names is 'all' else [name for name in var_names if var_names in adata.var_names]

    figsize = rcParams['figure.figsize']
    ncols = len(var_names)
    for i, gs in enumerate(pl.GridSpec(1, ncols, pl.figure(None, (figsize[0] * ncols, figsize[1]), dpi=100))):
        idx = np.where(data.var_names == var_names[i])[0][0]

        alpha, ut, st, _ = compute_dynamics(adata, idx)

        u = make_dense(data.layers['unspliced'][:, idx])
        s = make_dense(data.layers['spliced'][:, idx])

        pl.subplot(gs)

        pl.plot(t, alpha, label='alpha', linestyle='--', color='grey', linewidth=1)

        pl.plot(t, ut, label='unspliced', color='darkblue')
        pl.plot(t, st, label='spliced', color='red')

        pl.plot(t, u, color='darkblue', linewidth=.3)
        pl.plot(t, s, color='red', linewidth=.3)

        pl.xlim(0); pl.ylim(0); pl.legend()
        pl.xlabel('t'); pl.ylabel('counts')



from .preprocessing.utils import show_proportions, cleanup, set_initial_size, get_initial_size
from .preprocessing.neighbors import get_connectivities
from .preprocessing.moments import second_order_moments, second_order_moments_u

from .tools.utils import *
from .tools.rank_velocity_genes import get_mean_var
from .tools.run import convert_to_adata, convert_to_loom
from .tools.optimization import leastsq_NxN, leastsq_generalized, maximum_likelihood
from .tools.velocity_confidence import random_subsample
from .tools.velocity_graph import vals_to_csr

from .plotting.utils import is_categorical, clip, interpret_colorkey
from .plotting.velocity_embedding_grid import compute_velocity_on_grid

from .read_load import clean_obs_names, merge



from .. import settings
from .. import logging as logg
from .utils import make_dense
from .dynamical_model_utils import unspliced, spliced, derivatives, derivatives_o

import numpy as np
import matplotlib.pyplot as pl
from scipy.sparse import issparse


class DynamicsRecovery:
    def __init__(self, adata=None, gene=None, u=None, s=None, use_raw=False, reinitialize=True):
        # extract actual data
        if u is None or s is None:
            _layers = adata[:, gene].layers
            u = _layers['unspliced'] if use_raw or 'Mu' not in _layers.keys() else _layers['Mu']
            s = _layers['spliced'] if use_raw or 'Ms' not in _layers.keys() else _layers['Ms']
        self.s, self.u = make_dense(s).copy(), make_dense(u).copy()
        self.use_raw = use_raw

        # initialize
        self.u0, self.u0_ = None, None
        self.s0, self.s0_ = None, None
        self.ut, self.st = [], []

        self.alpha, self.beta, self.gamma, self.alpha_, self.t_,  = None, None, None, None, None
        self.t, self.tau, self.o = 0, 0, 0
        if 'true_t' in adata.obs.keys():  # ToDo: remove this later
            self.o = np.array(adata.obs.true_t < adata[:, gene].var.true_t_[0], dtype=int)

        self.weights, self.u_b, self.s_b = None, None, None
        self.loss, self.dalpha, self.dbeta, self.dgamma = [], [], [], []
        self.m_dalpha, self.m_dbeta, self.m_dgamma = [0], [0], [0]
        self.v_dalpha, self.v_dbeta, self.v_dgamma = [0], [0], [0]

        self.m_dpars = np.zeros((3, 1))
        self.v_dpars = np.zeros((3, 1))

        if reinitialize or 'fit_alpha' not in adata.var.keys():
            self.initialize()
        else:
            self.load_pars(adata, gene)

    def initialize(self, perc=95):
        u, s = self.u, self.s
        su_norm = s / np.clip(s.max(0), 1e-3, None) + u / np.clip(u.max(0), 1e-3, None)
        weights = su_norm >= np.percentile(su_norm, perc, axis=0)
        x = weights.multiply(s).tocsr() if issparse(weights) else weights * s
        y = weights.multiply(u).tocsr() if issparse(weights) else weights * u
        xx_ = x.multiply(x).sum(0) if issparse(x) else (x ** 2).sum(0)
        xy_ = x.multiply(y).sum(0) if issparse(x) else (x * y).sum(0)
        gamma = xy_ / xx_

        alpha_, beta = 0, 1
        alpha = u[u >= np.percentile(u, 95)].mean() * beta

        self.alpha, self.beta, self.gamma, self.alpha_ = alpha, beta, gamma, alpha_
        self.u0, self.u0_ = 0, alpha / beta * .8
        self.s0, self.s0_ = 0, alpha / gamma * .8

        self.update_state_dependent()

    def load_pars(self, adata, gene):
        idx = np.where(adata.var_names == gene)[0][0] if isinstance(gene, str) else gene
        self.alpha = adata.var['fit_alpha'][idx]
        self.beta = adata.var['fit_beta'][idx]
        self.gamma = adata.var['fit_gamma'][idx]
        self.t_ = adata.var['fit_t_'][idx]
        self.t = adata.layers['fit_t'][:, idx]

        self.u0, self.s0, self.alpha_ = 0, 0, 0
        self.u0_ = unspliced(self.t_, self.u0, self.alpha, self.beta)
        self.s0_ = spliced(self.t_, self.u0, self.s0, self.alpha, self.beta, self.gamma)
        self.update_state_dependent()

    def fit(self, n_iter=100, r=1e-3, inits=None, state_gd=True, b1=0.9, b2=0.999, eps=1e-8, method='adam'):
        for j in range(n_iter):
            self.update_vars(r=r, inits=inits, state_gd=state_gd, b1=b1, b2=b2, eps=eps, method=method)
            self.update_state_dependent()

    def update_state_dependent(self):
        def log(x):  # to avoid invalid values for log.
            return np.log(np.clip(x, 1e-6, 1 - 1e-6))

        u, s = self.u, self.s
        alpha, beta, gamma, alpha_ = self.alpha, self.beta, self.gamma, self.alpha_
        u0, s0, u0_, s0_ = self.u0, self.s0, self.u0_, self.s0_

        # assign states (on if above gamma fit, else off)
        if True:  # ToDo: remove later
            o = self.o
        else:
            o = np.array(u * s0_ - s * u0_ >= 0, dtype=int)

        # vectorize state-dependent vars
        alpha = alpha * o + alpha_ * (1 - o)
        u0 = u0 * o + u0_ * (1 - o)
        s0 = s0 * o + s0_ * (1 - o)

        # tau and t (explicitly given)
        u_inf = alpha / beta
        tau = - 1 / beta * log((u - u_inf) / (u0 - u_inf))
        t_ = np.max(tau * o)
        t = tau * o + (t_ + tau) * (1 - o)

        # find optimal t_
        #u_inf = alpha / beta
        #t_ = np.min(np.max(t * o), - 1 / beta * log((u_inf - u0_) / u_inf))

        #u0_ = unspliced(t_, 0, alpha, beta)
        #print(t_.round(2), u0_.round(2))

        # remove jumps in time assignment
        idx_ordered = np.argsort(t)
        t_ord = t[idx_ordered]
        dt_ord = t_ord - np.insert(t_ord[:-1], 0, 0)
        dt_ord = np.clip(dt_ord, None, 2 * np.percentile(dt_ord, 95))
        t[idx_ordered] = np.cumsum(dt_ord)

        # find optimal t_
        t_ = np.max(t * o)
        tau = t * o + (t - t_) * (1 - o)

        # estimated data via ODE
        ut = unspliced(tau, u0, alpha, beta)
        st = spliced(tau, s0, u0, alpha, beta, gamma)

        self.ut, self.st = ut, st
        self.t, self.tau, self.o, self.t_ = t.round(2), tau.round(2), o, t_

    def update_vars(self, r=1e-6, inits=None, state_gd=True, b1=0.9, b2=0.999, eps=1e-8, n_regions=5, method='adam'):
        inits = 'cont' if inits is None else inits
        state_gd = False if 'learn' in inits else state_gd
        if self.weights is None:
            self.uniform_weighting(n_regions=n_regions, perc=95)
        dalpha, dbeta, dgamma, dalpha_, du0, ds0, du0_, ds0_, loss = self.get_derivatives(state_gd)
        self.loss = np.append(self.loss, loss)
        self.dalpha = np.append(self.dalpha, dalpha)
        self.dbeta = np.append(self.dbeta, dbeta)
        self.dgamma = np.append(self.dgamma, dgamma)

        if method is 'adam':
            # update 1st and 2nd order gradient moments
            dpars = np.array([dalpha, dbeta, dgamma])
            m_dpars = b1 * self.m_dpars[:, -1] + (1 - b1) * dpars
            v_dpars = b2 * self.v_dpars[:, -1] + (1 - b2) * dpars**2

            self.m_dpars = np.c_[self.m_dpars, m_dpars]
            self.v_dpars = np.c_[self.v_dpars, v_dpars]

            # correct for bias
            t = len(self.m_dpars[0])
            m_dpars /= (1 - b1 ** t)
            v_dpars /= (1 - b2 ** t)

            # Adam parameter update
            self.alpha -= r * m_dpars[0] / (np.sqrt(v_dpars[0]) + eps)
            self.beta -= r * m_dpars[1] / (np.sqrt(v_dpars[1]) + eps)
            self.gamma -= r * m_dpars[2] / (np.sqrt(v_dpars[2]) + eps)
        else:
            self.alpha -= r * dalpha
            self.beta -= r * dbeta
            self.gamma -= r * dgamma

        if inits is 'ss':
            self.u0_ = np.array(self.alpha / self.beta)
            self.s0_ = np.array(self.alpha / self.gamma)
        elif inits is 'cont':
            self.u0_ = np.max(self.ut * self.o)
            self.s0_ = np.max(self.st * self.o)
        elif inits is 'learn':
            self.u0_ -= du0_ * r
            self.s0_ -= ds0_ * r
        elif inits is 'learn_all':
            self.alpha_ -= dalpha_ * r
            self.u0 -= du0 * r
            self.s0 -= ds0 * r
            self.u0_ -= du0_ * r
            self.s0_ -= ds0_ * r
        elif inits is 'lm':
            off = self.o == 0
            expu = np.exp(-self.beta * self.tau[off])
            exps = np.exp(-self.gamma * self.tau[off])
            self.u0_ = (expu * self.u[off]).sum() / (expu ** 2).sum()

            s = self.s[off] + self.beta * self.u0_ / (self.gamma - self.beta) * (exps - expu)
            self.s0_ = (exps * s).sum() / (exps ** 2).sum()

    def get_derivatives(self, state_gd=True):
        alpha, beta, gamma, alpha_ = self.alpha, self.beta, self.gamma, self.alpha_
        t, tau, o = self.t, self.tau, self.o
        u0, s0, u0_, s0_ = self.u0, self.s0, self.u0_, self.s0_
        diffU = np.multiply(np.array(self.ut - self.u), self.weights)  # TODO check if this actually are np arrays
        diffS = np.multiply(np.array(self.st - self.s), self.weights)

        if state_gd:
            dL_dalpha, dL_dbeta, dL_dgamma, dL_dalpha_, dL_du0, dL_ds0, dL_du0_, dL_ds0_ = \
                derivatives_o(alpha, beta, gamma, tau, o, diffU, diffS, alpha_, u0, s0, u0_, s0_)
        else:
            dL_dalpha, dL_dbeta, dL_dgamma, dL_dalpha_, dL_du0, dL_ds0, dL_du0_, dL_ds0_ = \
                derivatives(alpha, beta, gamma, tau, o, diffU, diffS, alpha_, u0, s0, u0_, s0_)

        loss = np.sqrt(np.sum(diffU ** 2 + diffS ** 2) / len(diffU))
        return dL_dalpha, dL_dbeta, dL_dgamma, dL_dalpha_, dL_du0, dL_ds0, dL_du0_, dL_ds0_, loss

    def uniform_weighting(self, n_regions=5, perc=95):
        from numpy import union1d as union
        from numpy import intersect1d as intersect
        u, s = self.u, self.s
        u_b = np.linspace(0, np.percentile(u, perc), n_regions)
        s_b = np.linspace(0, np.percentile(s, perc), n_regions)

        regions, weights = {}, np.ones(len(u))
        for i in range(n_regions):
            if i == 0:
                region = intersect(np.where(u < u_b[i + 1]), np.where(s < s_b[i + 1]))
            elif i < n_regions - 1:
                lower_cut = union(np.where(u > u_b[i]), np.where(s > s_b[i]))
                upper_cut = intersect(np.where(u < u_b[i + 1]), np.where(s < s_b[i + 1]))
                region = intersect(lower_cut, upper_cut)
            else:
                region = union(np.where(u > u_b[i]), np.where(s > s_b[i]))  # lower_cut for last region
            regions[i] = region
            if len(region) > 0:
                weights[region] = n_regions / len(region)
        # set weights accordingly such that each region has an equal overall contribution.
        self.weights = weights * len(u) / np.sum(weights)
        self.u_b, self.s_b = u_b, s_b

    def plot_regions(self):
        u, s, ut, st = self.u, self.s, self.ut, self.st
        u_b, s_b = self.u_b, self.s_b

        pl.figure(dpi=100)
        pl.scatter(s, u, color='grey')
        pl.xlim(0); pl.ylim(0); pl.xlabel('spliced'); pl.ylabel('unspliced')

        for i in range(len(s_b)):
            pl.plot([s_b[i], s_b[i], 0], [0, u_b[i], u_b[i]])

    def plot_derivatives(self):
        from .dynamical_model_utils import du_dalpha_o, du_dbeta_o, ds_dalpha_o, ds_dbeta_o, ds_dgamma_o
        alpha, beta, gamma, = self.alpha, self.beta, self.gamma
        t, tau, o = self.t, self.tau, self.o
        idx = np.argsort(t)
        t = t[idx]
        pl.plot(t, du_dalpha_o(beta, tau, o)[idx], label=r'$\partial u / \partial\alpha$')
        pl.plot(t, .2 * du_dbeta_o(alpha, beta, tau, o)[idx], label=r'$\partial u / \partial \beta$')
        pl.plot(t, ds_dalpha_o(beta, gamma, tau, o)[idx], label=r'$\partial s / \partial \alpha$')
        pl.plot(t, ds_dbeta_o(alpha, beta, gamma, tau, o)[idx], label=r'$\partial s / \partial \beta$')
        pl.plot(t, .2 * ds_dgamma_o(alpha, beta, gamma, tau, o)[idx], label=r'$\partial s / \partial \gamma$')

        pl.legend()
        pl.xlabel('t')


def read_pars(adata, pars_names=['alpha', 'beta', 'gamma', 't_'], key='fit'):
    pars = []
    for name in pars_names:
        pkey = key + '_' + name
        par = adata.var[pkey].values if pkey in adata.var.keys() else np.zeros(adata.n_vars) * np.nan
        pars.append(par)
    return pars


def write_pars(adata, pars, pars_names=['alpha', 'beta', 'gamma', 't_'], add_key='fit'):
    for i, name in enumerate(pars_names):
        adata.var[add_key + '_' + name] = pars[i]


def recover_dynamics(data, var_names='all', n_iter=100, learning_rate=1e-3, inits=None, state_gd=True, add_key='fit',
                     reinitialize=True, use_raw=False, copy=False, **kwargs):
    """Estimates velocities in a gene-specific manner

    Arguments
    ---------
    data: :class:`~anndata.AnnData`
        Annotated data matrix.

    Returns
    -------
    Returns or updates `adata`
    """
    adata = data.copy() if copy else data
    logg.info('recovering dynamics', r=True)

    idx = np.ones(adata.n_vars, dtype=bool) if var_names is 'all' else adata.var_names.isin(var_names)
    if 'velocity_genes' in var_names and 'velocity_genes' in adata.var.keys():
        idx = idx & np.array(adata.var.velocity_genes.values)
    idx = np.where(idx)[0]
    var_names = adata.var_names[idx]

    alpha, beta, gamma, t_ = read_pars(adata)
    T = adata.layers['fit_t'] if 'fit_t' in adata.layers.keys() else np.zeros(adata.shape) * np.nan
    L = np.zeros(shape=(n_iter, adata.n_vars)) * np.nan

    for i, gene in enumerate(var_names):
        drec = DynamicsRecovery(adata, gene, use_raw=use_raw, reinitialize=reinitialize)
        drec.fit(n_iter, learning_rate, inits=inits, state_gd=state_gd, **kwargs)

        i = idx[i]
        alpha[i], beta[i], gamma[i], t_[i] = drec.alpha, drec.beta, drec.gamma, drec.t_
        T[:, i] = drec.t
        L[:, i] = drec.loss

    write_pars(adata, [alpha, beta, gamma, t_])
    adata.layers['fit_t'] = T
    adata.varm['loss'] = L.T

    logg.info('    finished', time=True, end=' ' if settings.verbosity > 2 else '\n')
    logg.hint('added \n' 
              '    \'' + add_key + '_pars' + '\', fitted parameters for splicing dynamics (adata.var)')

    return adata if copy else None



from .. import settings
from .. import logging as logg
from .utils import not_yet_normalized, normalize_per_cell
from .neighbors import neighbors, get_connectivities, neighbors_to_be_recomputed

from scipy.sparse import csr_matrix
import numpy as np


def moments(data, n_neighbors=30, n_pcs=30, mode='connectivities', method='umap', metric='euclidean', use_rep=None,
            recurse_neighbors=False, renormalize=False, copy=False):
    """Computes moments for velocity estimation.

    Arguments
    ---------
    data: :class:`~anndata.AnnData`
        Annotated data matrix.
    n_neighbors: `int` (default: 30)
        Number of neighbors to use.
    n_pcs: `int` (default: 30)
        Number of principal components to use.
    mode: `'connectivities'` or `'distances'`  (default: `'connectivities'`)
        Distance metric to use for moment computation.
    renormalize: `bool` (default: `False`)
        Renormalize the moments by total counts per cell to its median.
    copy: `bool` (default: `False`)
        Return a copy instead of writing to adata.

    Returns
    -------
    Returns or updates `adata` with the attributes
    Ms: `.layers`
        dense matrix with first order moments of spliced counts.
    Mu: `.layers`
        dense matrix with first order moments of unspliced counts.
    """
    adata = data.copy() if copy else data

    if 'spliced' not in adata.layers.keys() or 'unspliced' not in adata.layers.keys():
        raise ValueError('Could not find spliced / unspliced counts.')
    if any([not_yet_normalized(adata.layers[layer]) for layer in {'spliced', 'unspliced'}]):
        normalize_per_cell(adata)
    if 'neighbors' not in adata.uns.keys() or neighbors_to_be_recomputed(adata, n_neighbors=n_neighbors):
        if use_rep is None: use_rep = 'X_pca'
        neighbors(adata, n_neighbors=n_neighbors, use_rep=use_rep, n_pcs=n_pcs, method=method, metric=metric)
    if mode not in adata.uns['neighbors']:
        raise ValueError('mode can only be \'connectivities\' or \'distances\'')

    logg.info('computing moments based on ' + str(mode), r=True)

    connectivities = get_connectivities(adata, mode, n_neighbors=n_neighbors, recurse_neighbors=recurse_neighbors)

    adata.layers['Ms'] = csr_matrix.dot(connectivities, csr_matrix(adata.layers['spliced'])).astype(np.float32).A
    adata.layers['Mu'] = csr_matrix.dot(connectivities, csr_matrix(adata.layers['unspliced'])).astype(np.float32).A
    if renormalize: normalize_per_cell(adata, layers={'Ms', 'Mu'}, enforce=True)

    logg.info('    finished', time=True, end=' ' if settings.verbosity > 2 else '\n')
    logg.hint(
        'added \n'
        '    \'Ms\' and \'Mu\', moments of spliced/unspliced abundances (adata.layers)')
    return adata if copy else None


def second_order_moments(adata, adjusted=False):
    """Computes second order moments for stochastic velocity estimation.

    Arguments
    ---------
    adata: `AnnData`
        Annotated data matrix.

    Returns
    -------
    Mss: Second order moments for spliced abundances
    Mus: Second order moments for spliced with unspliced abundances
    """
    if 'neighbors' not in adata.uns:
        raise ValueError('You need to run `pp.neighbors` first to compute a neighborhood graph.')

    connectivities = get_connectivities(adata)
    s, u = csr_matrix(adata.layers['spliced']), csr_matrix(adata.layers['unspliced'])
    Mss = csr_matrix.dot(connectivities, s.multiply(s)).astype(np.float32).A
    Mus = csr_matrix.dot(connectivities, s.multiply(u)).astype(np.float32).A
    if adjusted:
        Mss = 2 * Mss - adata.layers['Ms'].reshape(Mss.shape)
        Mus = 2 * Mus - adata.layers['Mu'].reshape(Mus.shape)
    return Mss, Mus


def second_order_moments_u(adata):
    """Computes second order moments for stochastic velocity estimation.

    Arguments
    ---------
    adata: `AnnData`
        Annotated data matrix.

    Returns
    -------
    Muu: Second order moments for unspliced abundances
    """
    if 'neighbors' not in adata.uns:
        raise ValueError('You need to run `pp.neighbors` first to compute a neighborhood graph.')

    connectivities = get_connectivities(adata)
    u = csr_matrix(adata.layers['unspliced'])
    Muu = csr_matrix.dot(connectivities, u.multiply(u)).astype(np.float32).A

    return Muu



"""Builtin Datasets.
"""
from .read_load import read, load
from .preprocessing.utils import cleanup
from anndata import AnnData
import numpy as np
import pandas as pd


def toy_data(n_obs):
    """
    Randomly samples from the Dentate Gyrus dataset.

    Arguments
    ---------
    n_obs: `int`
        Size of the sampled dataset

    Returns
    -------
    Returns `adata` object
    """

    """Random samples from Dentate Gyrus.
    """
    adata = dentategyrus()
    indices = np.random.choice(adata.n_obs, n_obs)
    adata = adata[indices]
    adata.obs_names = np.array(range(adata.n_obs), dtype='str')
    adata.var_names_make_unique()
    return adata


def dentategyrus():
    """Dentate Gyrus dataset from Hochgerner et al. (2018).

    Dentate gyrus is part of the hippocampus involved in learning, episodic memory formation and spatial coding.
    It is measured using 10X Genomics Chromium and described in Hochgerner et al. (2018).
    The data consists of 25,919 genes across 3,396 cells and provides several interesting characteristics.

    Returns
    -------
    Returns `adata` object
    """
    filename = 'data/DentateGyrus/10X43_1.loom'
    url = 'http://pklab.med.harvard.edu/velocyto/DG1/10X43_1.loom'
    adata = read(filename, backup_url=url, cleanup=True, sparse=True, cache=True)
    cleanup(adata, clean='all', keep={'spliced', 'unspliced', 'ambiguous'})

    url_louvain = 'https://github.com/theislab/scvelo_notebooks/raw/master/data/DentateGyrus/DG_clusters.npy'
    url_umap = 'https://github.com/theislab/scvelo_notebooks/raw/master/data/DentateGyrus/DG_umap.npy'

    adata.obs['clusters'] = load('./data/DentateGyrus/DG_clusters.npy', url_louvain)
    adata.obsm['X_umap'] = load('./data/DentateGyrus/DG_umap.npy', url_umap)

    adata.obs['clusters'] = pd.Categorical(adata.obs['clusters'])

    return adata


def forebrain():
    """Developing human forebrain.
    Forebrain tissue of a week 10 embryo, focusing on the glutamatergic neuronal lineage.

    Returns
    -------
    Returns `adata` object
    """
    filename = 'data/ForebrainGlut/hgForebrainGlut.loom'
    url = 'http://pklab.med.harvard.edu/velocyto/hgForebrainGlut/hgForebrainGlut.loom'
    adata = read(filename, backup_url=url, cleanup=True, sparse=True, cache=True)
    adata.var_names_make_unique()
    return adata


def simulation(n_obs=300, n_vars=4, alpha=None, beta=None, gamma=None, alpha_=None, t_max=None,
               noise_model='normal', noise_level=1):
    """Simulation of mRNA metabolism with transcription, splicing and degradation

    Returns
    -------
    Returns `adata` object
    """
    from .tools.dynamical_model_utils import unspliced, spliced, vectorize

    def draw_poisson(n):
        from random import uniform  # draw from poisson
        t = np.cumsum([- .1 * np.log(uniform(0, 1)) for _ in range(n - 1)])
        return np.insert(t, 0, 0)  # prepend t0=0

    def simulate_dynamics(tau, alpha, beta, gamma, u0, s0, noise_model, noise_level):
        ut = unspliced(tau, u0, alpha, beta)
        st = spliced(tau, s0, u0, alpha, beta, gamma)

        if noise_model is 'normal':  # add noise
            ut += np.random.normal(scale=noise_level * np.percentile(ut, 99) / 10, size=len(ut))
            st += np.random.normal(scale=noise_level * np.percentile(st, 99) / 10, size=len(st))
        ut, st = np.clip(ut, 0, None), np.clip(st, 0, None)
        return ut, st

    def simulate_gillespie(alpha, beta, gamma):
        # update rules: transcription (u+1,s), splicing (u-1,s+1), degradation (u,s-1), nothing (u,s)
        update_rule = np.array([[1, 0], [-1, 1], [0, -1], [0, 0]])

        def update(props):
            if np.sum(props) > 0:
                props /= np.sum(props)
            p_cumsum = props.cumsum()
            p = np.random.rand()
            i = 0
            while p > p_cumsum[i]:
                i += 1
            return update_rule[i]

        u, s = np.zeros(len(alpha)), np.zeros(len(alpha))
        for i, alpha_i in enumerate(alpha):
            u_, s_ = (u[i - 1], s[i - 1]) if i > 0 else (0, 0)
            du, ds = update(props=np.array([alpha_i, beta * u_, gamma * s_]))
            u[i], s[i] = (u_ + du, s_ + ds)
        return u, s

    alpha = 5 if alpha is None else alpha
    beta = .5 if beta is None else beta
    gamma = .3 if gamma is None else gamma
    alpha_ = 0 if alpha_ is None else alpha_

    t = draw_poisson(n_obs)
    if t_max is not None:
        t *= t_max / np.max(t)
    t_max = np.max(t)

    # switching time point obtained as fraction of t_max rounded down
    t_ = np.array([np.max(t[t < t_i * t_max]) for t_i in [.4, .7, 1, .1]])

    U = np.zeros(shape=(len(t), n_vars))
    S = np.zeros(shape=(len(t), n_vars))
    for i in range(n_vars):
        tau, alpha_vec, u0_vec, s0_vec = vectorize(t, t_[i], alpha, beta, gamma, alpha_=alpha_, u0=0, s0=0)
        if noise_model is 'gillespie':
            U[:, i], S[:, i] = simulate_gillespie(alpha_vec, beta, gamma)
        else:
            U[:, i], S[:, i] = simulate_dynamics(tau, alpha_vec, beta, gamma, u0_vec, s0_vec, noise_model, noise_level)

    obs = {'true_t': t.round(2)}
    var = {'true_t_': t_[:n_vars],
           'true_alpha': np.ones(n_vars) * alpha,
           'true_beta': np.ones(n_vars) * beta,
           'true_gamma': np.ones(n_vars) * gamma}
    layers = {'unspliced': U, 'spliced': S}

    return AnnData(S, obs, var, layers=layers)



from setuptools import setup, find_packages
import numpy as np
from version import __version__

setup(
    name="scvelo",
    version=__version__,
    install_requires=['scanpy>=1.3.3', 'anndata>=0.6.18', 'loompy>=2.0.12', 'numpy>=1.14', 'pandas>=0.23.0',
                      'scipy>=1.0', 'scikit-learn>=0.19.1', 'matplotlib>=2.2', 'setuptools'],
    packages=find_packages(),
    include_dirs=[np.get_include()],
    author="Volker Bergen",
    author_email="volker.bergen@helmholtz-muenchen.de",
    description='Stochastic RNA velocity for inferring single cell dynamics',
    license='BSD',
    url="https://github.com/theislab/scvelo",
    download_url=f"https://github.com/theislab/scvelo",
    keywords=["RNAseq", "singlecell", "stochastic", "velocity", "transcriptomics"]
    )



import numpy as np
import matplotlib.pyplot as pl
from matplotlib import rcParams
from matplotlib.colors import ColorConverter
from pandas import unique
from scipy.sparse import issparse
from .utils import is_categorical, interpret_colorkey, default_basis, default_size, get_components, savefig_or_show, \
    default_color, make_unique_list, set_colorbar, default_color_map, set_label, set_title


def heatmap(adata, var_names, groups=None, groupby=None, annotations=None, use_raw=False, layers=['X'], color_map=None,
            color_map_anno=None, colorbar=True, row_width=None, xlabel=None, title=None, figsize=None, dpi=None,
            show=True, save=None, ax=None, **kwargs):

    """\
    Plot pseudotimeseries for genes as heatmap.

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    var_names: `str`,  list of `str`
        Names of variables to use for the plot.
    groups: `str`,  list of `str` or `None` (default: `None`)
        Groups selected to plot. Must be an element of adata.obs[groupby].
    groupby: `str` or `None` (default: `None`)
        Key in adata.obs. Indicates how to group the plot.
    annotations: `str`,  list of `str` or `None` (default: `None`)
        Key in adata.obs. Annotations are plotted in the last row.
    use_raw: `bool` (default: `False`)
        If true, moments are used instead of raw data.
    layers: `str`,  list of `str` or `None` (default: `['X']`)
        Selected layers.
    color_map: `str`,  list of `str` or `None` (default: `None`)
        String denoting matplotlib color map for the heat map.
        There must be one list entry for each layer.
    color_map_anno: `str`,  list of `str` or `None` (default: `None`)
        String denoting matplotlib color map for the annotations.
        There must be one list entry for each annotation.
    colorbar: `bool` (default: `True`)
        If True, a colormap for each layer is added on the right bottom corner.
    row_width: `float` (default: `None`)
        Constant width of all rows.
    xlabel:
        Label for the x-axis.
    title: `str` or `None` (default: `None`)
        Main plot title.
    figsize: tuple (default: `(7,5)`)
        Figure size.
    dpi: `int` (default: 80)
        Figure dpi.
    show: `bool`, optional (default: `None`)
        Show the plot, do not return axis.
    save: `bool` or `str`, optional (default: `None`)
        If `True` or a `str`, save the figure. A string is appended to the default filename.
        Infer the filetype if ending on {'.pdf', '.png', '.svg'}.
    ax: `matplotlib.Axes`, optional (default: `None`)
        A matplotlib axes object. Only works if plotting a single component.

    Returns
    -------
        If `show==False` a `matplotlib.Axis`
    """

    # catch
    if 'velocity_pseudotime' not in adata.obs.keys():
        raise ValueError(
            'A function requires computation of the pseudotime'
            'for ordering at single-cell resolution')
    if annotations is None:
        annotations = []
    if isinstance(var_names, str):
        var_names = [var_names]
    if len(var_names) == 0:
        var_names = np.arange(adata.X.shape[1])
    if var_names.ndim == 2:
        var_names = var_names[:, 0]
    var_names = [name for name in var_names if name in adata.var_names]
    if len(var_names) == 0:
        raise ValueError(
            'The specified var_names are all not'
            'contained in the adata.var_names.')

    if layers is None:
        layers = ['X']
    if isinstance(layers, str):
        layers = [layers]
    layers = [layer for layer in layers if layer in adata.layers.keys() or layer == 'X']
    if len(layers) == 0:
        raise ValueError(
            'The selected layers are not contained'
            'in adata.layers.keys().')
    if not use_raw:
        layers = np.array(layers)
        if 'X' in layers: layers[np.array([layer == 'X' for layer in layers])] = 'Ms'
        if 'spliced' in layers: layers[np.array([layer == 'spliced' for layer in layers])] = 'Ms'
        if 'unspliced' in layers: layers[np.array([layer == 'unspliced' for layer in layers])] = 'Ms'
        layers = list(layers)
    if 'Ms' in layers and 'Ms' not in adata.layers.keys():
        raise ValueError(
            'Moments have to be computed before'
            'using this plot function.')
    if 'Mu' in layers and 'Mu' not in adata.layers.keys():
        raise ValueError(
            'Moments have to be computed before'
            'using this plot function.')
    layers = unique(layers)

    # Number of rows to plot
    tot_len = len(var_names) * len(layers) + len(annotations)

    # init main figure
    figsize = rcParams['figure.figsize'] if figsize is None else figsize
    if row_width is not None: figsize[1] = row_width * tot_len
    ax = pl.figure(figsize=figsize, dpi=dpi).gca() if ax is None else ax
    ax.set_yticks([])
    ax.set_xticks([])

    # groups bar
    ax_bounds = ax.get_position().bounds
    if groupby is not None:
        # catch
        if groupby not in adata.obs_keys():
            raise ValueError(
                'The selected groupby is not contained'
                'in adata.obs_keys().')
        if groups is None:  # Then use everything of that obs
            groups = unique(adata.obs.clusters.values)

        imlist = []

        for igroup, group in enumerate(groups):
            for ivar, var in enumerate(var_names):
                for ilayer, layer in enumerate(layers):
                    groups_axis = pl.axes([ax_bounds[0] + igroup * ax_bounds[2] / len(groups),
                                           ax_bounds[1] + ax_bounds[3] * (
                                                       tot_len - ivar * len(layers) - ilayer - 1) / tot_len,
                                           ax_bounds[2] / len(groups),
                                           (ax_bounds[3] - ax_bounds[3] / tot_len * len(annotations)) / (
                                                       len(var_names) * len(layers))])

                    # Get data to fill and reshape
                    dat = adata[:, var]

                    idx_group = [adata.obs[groupby] == group]
                    idx_group = np.array(idx_group[0].tolist())
                    idx_var = [vn in var_names for vn in adata.var_names]
                    idx_pt = np.array(adata.obs.velocity_pseudotime).argsort()
                    idx_pt = idx_pt[np.array(np.isnan(np.array(dat.obs.velocity_pseudotime)[idx_pt]) == False)]

                    if layer == 'X':
                        laydat = dat.X
                    else:
                        laydat = dat.layers[layer]

                    t1, t2, t3 = idx_group, idx_var, idx_pt
                    t1 = t1[t3]
                    # laydat = laydat[:, t2]  # select vars
                    laydat = laydat[t3]
                    laydat = laydat[t1]  # select ordered groups

                    if issparse(laydat):
                        laydat = laydat.A

                    # transpose X for ordering in direction var_names: up->downwards
                    laydat = laydat.T[::-1]
                    laydat = laydat.reshape((1, len(laydat)))  # ensure 1dimty

                    # plot
                    im = groups_axis.imshow(laydat, aspect='auto', interpolation="nearest", cmap=color_map[ilayer])

                    # Frames
                    if ilayer == 0:
                        groups_axis.spines['bottom'].set_visible(False)
                    elif ilayer == len(layer) - 1:
                        groups_axis.spines['top'].set_visible(False)
                    else:
                        groups_axis.spines['top'].set_visible(False)
                        groups_axis.spines['bottom'].set_visible(False)

                    # Further visuals
                    if igroup == 0:
                        if colorbar:
                            if len(layers) % 2 == 0:
                                if ilayer == len(layers) / 2 - 1:
                                    pl.yticks([0.5], [var])
                                else:
                                    groups_axis.set_yticks([])
                            else:
                                if ilayer == (len(layers) - 1) / 2:
                                    pl.yticks([0], [var])
                                else:
                                    groups_axis.set_yticks([])
                        else:
                            pl.yticks([0], [layer + ' ' + var])
                    else:
                        groups_axis.set_yticks([])

                    groups_axis.set_xticks([])
                    if ilayer == 0 and ivar == 0:
                        groups_axis.set_title(str(group))
                    groups_axis.grid(False)

                    # handle needed as mappable for colorbar
                    if igroup == len(groups) - 1:
                        imlist.append(im)

            # further annotations for each group
            if annotations is not None:
                for ianno, anno in enumerate(annotations):
                    anno_axis = pl.axes([ax_bounds[0] + igroup * ax_bounds[2] / len(groups),
                                         ax_bounds[1] + ax_bounds[3] / tot_len * (len(annotations) - ianno - 1),
                                         ax_bounds[2] / len(groups),
                                         ax_bounds[3] / tot_len])
                    if is_categorical(adata, anno):
                        colo = interpret_colorkey(adata, anno)[t3][t1]
                        colo.reshape(1, len(colo))
                        mapper = np.vectorize(ColorConverter.to_rgb)
                        a = mapper(colo)
                        a = np.array(a).T
                        Y = a.reshape(1, len(colo), 3)
                    else:
                        Y = np.array(interpret_colorkey(adata, anno))[t3][t1]
                        Y = Y.reshape(1, len(Y))
                    img = anno_axis.imshow(Y, aspect='auto',
                                           interpolation='nearest', cmap=color_map_anno)
                    if igroup == 0:
                        anno_axis.set_yticklabels(['', anno, ''])  # , fontsize=ytick_fontsize)
                        anno_axis.tick_params(axis='both', which='both', length=0)
                    else:
                        anno_axis.set_yticklabels([])
                        anno_axis.set_yticks([])
                    anno_axis.set_xticks([])
                    anno_axis.set_xticklabels([])
                    anno_axis.grid(False)
                    pl.ylim([.5, -.5])  # center ticks

    else:  # groupby is False
        imlist = []
        for ivar, var in enumerate(var_names):
            for ilayer, layer in enumerate(layers):
                ax_bounds = ax.get_position().bounds
                groups_axis = pl.axes([ax_bounds[0],
                                       ax_bounds[1] + ax_bounds[3] * (
                                                   tot_len - ivar * len(layers) - ilayer - 1) / tot_len,
                                       ax_bounds[2],
                                       (ax_bounds[3] - ax_bounds[3] / tot_len * len(annotations)) / (
                                                   len(var_names) * len(layers))])
                # Get data to fill
                dat = adata[:, var]
                idx = np.array(dat.obs.velocity_pseudotime).argsort()
                idx = idx[np.array(np.isnan(np.array(dat.obs.velocity_pseudotime)[idx]) == False)]

                if layer == 'X':
                    laydat = dat.X
                else:
                    laydat = dat.layers[layer]
                laydat = laydat[idx]
                if issparse(laydat):
                    laydat = laydat.A

                # transpose X for ordering in direction var_names: up->downwards
                laydat = laydat.T[::-1]
                laydat = laydat.reshape((1, len(laydat)))

                # plot
                im = groups_axis.imshow(laydat, aspect='auto', interpolation="nearest", cmap=color_map[ilayer])
                imlist.append(im)

                # Frames
                if ilayer == 0:
                    groups_axis.spines['bottom'].set_visible(False)
                elif ilayer == len(layer) - 1:
                    groups_axis.spines['top'].set_visible(False)
                else:
                    groups_axis.spines['top'].set_visible(False)
                    groups_axis.spines['bottom'].set_visible(False)

                # Further visuals
                groups_axis.set_xticks([])
                groups_axis.grid(False)
                pl.ylim([.5, -.5])  # center
                if colorbar:
                    if len(layers) % 2 == 0:
                        if ilayer == len(layers) / 2 - 1:
                            pl.yticks([0.5], [var])
                        else:
                            groups_axis.set_yticks([])
                    else:
                        if ilayer == (len(layers) - 1) / 2:
                            pl.yticks([0], [var])
                        else:
                            groups_axis.set_yticks([])
                else:
                    pl.yticks([0], [layer + ' ' + var])

        # further annotations bars
        if annotations is not None:
            for ianno, anno in enumerate(annotations):
                anno_axis = pl.axes([ax_bounds[0],
                                     ax_bounds[1] + ax_bounds[3] / tot_len * (len(annotations) - ianno - 1),
                                     ax_bounds[2],
                                     ax_bounds[3] / tot_len])
                dat = adata[:, var_names]
                if is_categorical(dat, anno):
                    colo = interpret_colorkey(dat, anno)[idx]
                    colo.reshape(1, len(colo))
                    mapper = np.vectorize(ColorConverter.to_rgb)
                    a = mapper(colo)
                    a = np.array(a).T
                    Y = a.reshape(1, len(idx), 3)
                else:
                    Y = np.array(interpret_colorkey(dat, anno)[idx]).reshape(1, len(idx))
                img = anno_axis.imshow(Y, aspect='auto', interpolation='nearest', cmap=color_map_anno)

                anno_axis.set_yticklabels(['', anno, ''])  # , fontsize=ytick_fontsize)
                anno_axis.tick_params(axis='both', which='both', length=0)
                anno_axis.grid(False)
                anno_axis.set_xticks([])
                anno_axis.set_xticklabels([])
                pl.ylim([-.5, +.5])

    # Colorbar
    if colorbar:
        if len(layers) > 1:
            # I must admit, this part is chaotic
            for ilayer, layer in enumerate(layers):
                w = 0.015 * 10 / figsize[0]  # 0.02 * ax_bounds[2]
                x = ax_bounds[0] + ax_bounds[2] * 0.99 + 1.5 * w + w * 1.2 * ilayer
                y = ax_bounds[1]
                h = ax_bounds[3] * .3
                cbaxes = pl.axes([x, y, w, h])
                cb = pl.colorbar(mappable=imlist[ilayer], cax=cbaxes)
                pl.text(x - 40 * w, y + h * 4, layer, rotation=45, horizontalalignment='left',
                        verticalalignment='bottom')
                if ilayer == len(layers) - 1:
                    ext = abs(cb.vmin - cb.vmax)
                    cb.set_ticks([cb.vmin + 0.07 * ext, cb.vmax - 0.07 * ext])
                    cb.ax.set_yticklabels(['Low', 'High'])  # vertical colorbar
                else:
                    cb.set_ticks([])
        else:
            cbaxes = pl.axes([ax_bounds[0] + ax_bounds[2] + .01,
                              ax_bounds[1],
                              0.02,
                              ax_bounds[3] * .3])
            cb = pl.colorbar(mappable=im, cax=cbaxes)
            cb.set_ticks([cb.vmin, cb.vmax])
            cb.ax.set_yticklabels(['Low', 'High'])

    if xlabel is None: xlabel = 'velocity' + ' ' + 'pseudotime'
    if title is not None: ax.set_title(title, pad=30)
    if len(annotations) == 0:
        ax.set_xlabel(xlabel)
        ax.xaxis.labelpad = 20

    # set_label(xlabel, None, fontsize, basis)
    # set_title(title, None, None, fontsize)
    # ax = update_axes(ax, fontsize)

    savefig_or_show('heatmap', dpi=dpi, save=save, show=show)
    if not show: return ax